import json
import os
import time
from pathlib import Path

import pandas as pd
import pytest

from indico.client import IndicoClient
from indico.errors import IndicoRequestError
from indico.filters import DatasetFilter
from indico.queries.datasets import (
    AddFiles,
    CreateDataset,
    CreateEmptyDataset,
    DeleteDataset,
    GetDataset,
    GetDatasetFileStatus,
    ListDatasets,
)
from indico.queries.export import CreateExport, DownloadExport
from indico.types.dataset import (
    Dataset,
    OcrEngine,
    OmnipageOcrOptionsInput,
    ReadApiOcrOptionsInput,
    TableReadOrder,
)
from tests.integration.data.datasets import airlines_dataset  # noqa: F401


def test_create_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"

    response = client.call(
        CreateDataset(
            name=f"AirlineComplaints-test-{int(time.time())}", files=[dataset_filepath]
        )
    )

    assert isinstance(response, Dataset)
    assert response.status == "COMPLETE"
    assert isinstance(response.id, int)


def test_get_datasets(indico, airlines_dataset):
    client = IndicoClient()
    dataset = client.call(GetDataset(id=airlines_dataset.id))

    assert isinstance(dataset, Dataset)
    assert dataset.id == airlines_dataset.id


def test_get_dataset_with_bad_id(indico):
    client = IndicoClient()

    with pytest.raises(IndicoRequestError):
        dataset = client.call(GetDataset(id=10000000))


def test_get_dataset_file_status(indico, airlines_dataset):
    client = IndicoClient()
    dataset = client.call(GetDatasetFileStatus(id=airlines_dataset.id))

    assert isinstance(dataset, Dataset)
    assert dataset.id == airlines_dataset.id
    assert len(dataset.files) > 0
    assert dataset.files[0].status != None


def test_list_datasets(indico, airlines_dataset):
    client = IndicoClient()
    datasets = client.call(ListDatasets(limit=1))

    assert isinstance(datasets, list)
    assert len(datasets) == 1
    assert type(datasets[0]) == Dataset


def test_list_datasets_filtered(indico, airlines_dataset):
    client = IndicoClient()
    datasets = client.call(
        ListDatasets(filters=DatasetFilter(name=f"{time.time()}_bananas"))
    )

    assert isinstance(datasets, list)
    assert len(datasets) == 0

    # happy path
    datasets = client.call(
        ListDatasets(filters=DatasetFilter(name=airlines_dataset.name))
    )

    assert isinstance(datasets, list)
    assert len(datasets) == 1
    assert type(datasets[0]) == Dataset
    assert datasets[0].name == airlines_dataset.name


def test_images(indico):
    client = IndicoClient()

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/dog_vs_cats_small.csv"
    response = client.call(
        CreateDataset(
            name=f"image-dataset-test-{int(time.time())}",
            files=dataset_filepath,
            from_local_images=True,
        )
    )
    assert isinstance(response, Dataset)
    assert response.status == "COMPLETE"
    assert isinstance(response.id, int)


def test_images_batch(indico):
    client = IndicoClient()

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/dog_vs_cats_small.csv"
    response = client.call(
        CreateDataset(
            name=f"image-dataset-test-{int(time.time())}",
            files=dataset_filepath,
            from_local_images=True,
            batch_size=10,
        )
    )
    assert isinstance(response, Dataset)
    assert response.status == "COMPLETE"
    assert isinstance(response.id, int)


def test_upload_pdf_dataset_batch(indico):
    client = IndicoClient()
    file_names = ["mock.pdf", "mock_2.pdf", "mock_3.pdf"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]
    dataset = client.call(
        CreateDataset(
            name=f"pdf-dataset-test-{int(time.time())}",
            files=dataset_filepaths,
            batch_size=1,
        )
    )
    assert isinstance(dataset, Dataset)
    assert dataset.status == "COMPLETE"
    assert isinstance(dataset.id, int)
    assert dataset.datacolumns[0].name == "text"
    response = client.call(GetDatasetFileStatus(id=dataset.id))
    for datafile in response.files:
        assert datafile.status == "PROCESSED"
        assert datafile.file_type == "PDF"


def test_upload_pdf_interrupt(indico):
    client = IndicoClient()

    client._http.request_session.cookies.clear()
    file_names = ["mock.pdf", "mock_2.pdf", "mock_3.pdf"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    dataset = client.call(
        CreateDataset(
            name=f"pdf-dataset-test-{int(time.time())}",
            files=dataset_filepaths,
            batch_size=1,
        )
    )


def test_delete_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"

    dataset = client.call(
        CreateDataset(
            name=f"AirlineComplaints-test-{int(time.time())}", files=[dataset_filepath]
        )
    )

    assert isinstance(dataset, Dataset)
    assert dataset.status == "COMPLETE"
    assert isinstance(dataset.id, int)

    success = client.call(DeleteDataset(id=dataset.id))
    assert success == True

    with pytest.raises(IndicoRequestError):
        dataset = client.call(GetDataset(id=dataset.id))


def _dataset_complete(dataset):
    for df in dataset.files:
        assert df.status == "PROCESSED"

    assert dataset.status == "COMPLETE"


@pytest.mark.ocr("omnipage")
def test_create_with_options(indico):
    client = IndicoClient()
    config: OmnipageOcrOptionsInput = {
        "auto_rotate": True,
        "single_column": True,
        "upscale_images": True,
        "languages": ["ENG", "FIN"],
        "force_render": False,
        "native_layout": False,
        "native_pdf": False,
        "table_read_order": TableReadOrder.ROW,
    }
    dataset = client.call(
        CreateEmptyDataset(
            name=f"dataset-{int(time.time())}",
            ocr_engine=OcrEngine.OMNIPAGE,
            omnipage_ocr_options=config,
        )
    )


@pytest.mark.ocr("readapi")
def test_create_with_options_readapi(indico):
    client = IndicoClient()
    config: ReadApiOcrOptionsInput = {
        "auto_rotate": True,
        "single_column": False,
        "upscale_images": True,
        "languages": ["AUTO"],
    }
    dataset = client.call(
        CreateEmptyDataset(
            name=f"dataset-{int(time.time())}",
            ocr_engine=OcrEngine.READAPI,
            readapi_ocr_options=config,
        )
    )


@pytest.mark.ocr("readapi_v2")
def test_create_from_files_with_readapiv2(indico):
    client = IndicoClient()
    dataset = client.call(
        CreateEmptyDataset(
            name=f"dataset-{int(time.time())}", ocr_engine=OcrEngine.READAPI_V2
        )
    )
    file_names = ["us_doi.tiff", "mock.pdf"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepaths[0]], autoprocess=True)
    )
    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepaths[1]], autoprocess=True)
    )

    _dataset_complete(dataset)


def test_create_from_files_document(indico):
    client = IndicoClient()
    dataset = client.call(CreateEmptyDataset(name=f"dataset-{int(time.time())}"))
    file_names = ["us_doi.tiff", "mock.pdf"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepaths[0]], autoprocess=True)
    )
    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepaths[1]], autoprocess=True)
    )

    _dataset_complete(dataset)


def test_create_from_files_image(indico):
    client = IndicoClient()
    dataset = client.call(
        CreateEmptyDataset(name=f"dataset-{int(time.time())}", dataset_type="IMAGE")
    )
    file_names = ["1.jpg", "2.jpg", "3.jpg"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=dataset_filepaths, autoprocess=True)
    )

    file_names = ["4.jpg", "5.jpg"]
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=dataset_filepaths, autoprocess=True)
    )
    _dataset_complete(dataset)


def test_create_from_csv(indico):
    client = IndicoClient()
    dataset = client.call(CreateEmptyDataset(name=f"dataset-{int(time.time())}"))
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"
    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepath], autoprocess=True)
    )
    _dataset_complete(dataset)
    full_dataset = dataset = client.call(GetDataset(id=dataset.id))
    export = client.call(
        CreateExport(
            dataset_id=full_dataset.id,
            labelset_id=full_dataset.labelsets[0].id,
            wait=True,
        )
    )

    exported_data = client.call(DownloadExport(export.id))
    exported_data.to_csv("test.csv")
    baseline_data = pd.read_csv(dataset_filepath)
    for col in baseline_data.columns:
        assert col in exported_data.columns
    assert (baseline_data["Text"] == exported_data["Text"]).all()
    export_labels = exported_data["Target_1"].apply(
        lambda x: json.loads(x)["targets"][0]["label"]
    )
    assert (baseline_data["Target_1"].str.strip() == export_labels).all()


def test_create_from_csv_image_urls(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/image_link_small.csv"
    dataset = client.call(
        CreateEmptyDataset(name=f"dataset-{int(time.time())}", dataset_type="IMAGE")
    )
    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepath], autoprocess=True)
    )

    _dataset_complete(dataset)


def test_create_from_csv_image_urls_with_broken(indico):
    client = IndicoClient()
    dataset_filepath = (
        str(Path(__file__).parents[1]) + "/data/image_link_small_with_broken.csv"
    )
    dataset = client.call(
        CreateEmptyDataset(name=f"dataset-{int(time.time())}", dataset_type="IMAGE")
    )
    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepath], autoprocess=True)
    )

    failed = 0
    for df in dataset.files:
        if df.status == "FAILED":
            assert df.failure_type == "DOWNLOAD"
            failed += 1
        else:
            assert df.status == "PROCESSED"
    assert failed == 1


def test_create_from_csv_doc_urls(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/pdf_links.csv"
    dataset = client.call(CreateEmptyDataset(name=f"dataset-{int(time.time())}"))
    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepath], autoprocess=True)
    )

    _dataset_complete(dataset)


def test_csv_incompat_columns(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/pdf_links.csv"
    dataset = client.call(CreateEmptyDataset(name=f"dataset-{int(time.time())}"))
    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepath], autoprocess=True)
    )

    for f in dataset.files:
        assert f.status == "PROCESSED"

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"
    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[dataset_filepath], autoprocess=True)
    )

    failed = 0
    for df in dataset.files:
        if df.status == "FAILED":
            assert df.failure_type == "INCOMPATIBLE_CSV_COLUMNS"
            failed += 1
        else:
            assert df.status == "PROCESSED"

    assert dataset.status == "COMPLETE"


def test_bad_csv_create_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/classification.csv"
    dataset = client.call(
        CreateDataset(
            name=f"dataset-{int(time.time())}", files=[dataset_filepath], wait=True
        ),
    )

    assert dataset.status == "CREATING"
    dataset = client.call(GetDatasetFileStatus(id=dataset.id))
    assert all([f.status == "FAILED" for f in dataset.files])


@pytest.mark.ocr("readapi")
def test_create_with_email_options_readapi(indico):
    client = IndicoClient()
    readapi_config: ReadApiOcrOptionsInput = {
        "auto_rotate": True,
        "single_column": False,
        "upscale_images": True,
        "languages": ["AUTO"],
    }
    email_config = {
        "include_sections": {"header": True, "body": True, "attachments": True},
        "unpack": True,
    }
    dataset = client.call(
        CreateEmptyDataset(
            name=f"dataset-{int(time.time())}",
            ocr_engine=OcrEngine.READAPI,
            readapi_ocr_options=readapi_config,
            email_options=email_config,
        )
    )



import os

from indico.client import IndicoClient
from indico.queries import (
    AddExchangeIntegration,
    StartIntegration,
    DeleteIntegration,
    PauseIntegration,
)

from indico.types import Workflow, Integration, ModelGroup
from tests.integration.data.datasets import (
    airlines_workflow,
    airlines_dataset,
    org_annotate_workflow,
    org_annotate_dataset,
    org_annotate_exchange_integration,
    org_annotate_model_group,
    exchange_integration_to_delete,
    started_exchange_integration
)

def test_add_integration(airlines_workflow: Workflow):
    client = IndicoClient()
    creds = {
        "clientId": os.getenv("EXCH_CLIENT_ID"),
        "clientSecret": os.getenv("EXCH_CLIENT_SECRET"),
        "tenantId": os.getenv("EXCH_TENANT_ID"),
    }

    config = {"userId": os.getenv("EXCH_USER_ID"), "folderId": "mailFolders('inbox')"}

    integ: Integration = client.call(
        AddExchangeIntegration(
            workflow_id=airlines_workflow.id, config=config, credentials=creds
        )
    )

    assert integ.workflow_id == airlines_workflow.id
    assert integ.config.folder_name == "Inbox"


def test_start_integration(
    org_annotate_exchange_integration: Integration,
    org_annotate_model_group: ModelGroup,
    org_annotate_workflow: Workflow,
):
    integ = org_annotate_exchange_integration
    assert not integ.enabled
    client = IndicoClient()
    resp = client.call(
        StartIntegration(
            integration_id=integ.id
        )
    )
    assert resp["startWorkflowIntegration"]["success"]
    

def test_delete_integration(exchange_integration_to_delete: Integration):
    integ = exchange_integration_to_delete
    client = IndicoClient()
    resp = client.call(
        DeleteIntegration(
            integration_id=integ.id
        )
    )
    assert resp["deleteWorkflowIntegration"]["success"]


def test_pause_integration(started_exchange_integration: Integration):
    integ = started_exchange_integration
    client = IndicoClient()
    resp = client.call(
        PauseIntegration(
            integration_id=integ.id
        )
    )
    assert resp["pauseWorkflowIntegration"]["success"]



from datetime import datetime
from typing import List

from indico import IndicoClient
from indico.queries.document_report import GetDocumentReport
from indico.types.document_report import DocumentReport
from indico.filters import DocumentReportFilter


def test_fetch_docs(indico):
    """
    NOTE: this test will fail if the cluster hasn't yet processed any submissions
    """
    client = IndicoClient()

    document_report: List[DocumentReport] = client.call(GetDocumentReport())
    assert document_report is not None
    assert len(document_report) > 0
    input_files = [f for dr in document_report for f in dr.input_files]
    num_pages_check = [isinstance(f.num_pages, int) for f in input_files if f.num_pages]
    if num_pages_check:
        assert all(num_pages_check)
    file_size_check = [isinstance(f.file_size, int) for f in input_files if f.file_size]
    if file_size_check:
        assert all(file_size_check)
    assert all([isinstance(f.id, int) for f in input_files])
    assert all([isinstance(f.submission_id, int) for f in input_files])
    assert [f.filename for f in input_files]


def test_fetch_docs_limit(indico):
    """
    NOTE: this test will fail if the cluster hasn't yet processed any submissions
    """
    client = IndicoClient()
    filter_opts = DocumentReportFilter(
        created_at_start_date=datetime(2021, 7, 1), created_at_end_date=datetime.now()
    )
    document_report: List[DocumentReport] = client.call(
        GetDocumentReport(filters=filter_opts, limit=5)
    )
    assert document_report is not None
    assert len(document_report) > 0
    assert len(document_report) <= 5


def test_pagination(indico):
    """
    NOTE: this test will fail if the cluster hasn't yet processed any submissions
    """
    client = IndicoClient()
    document_report: List[DocumentReport] = []
    num_pages_to_check = 5
    for i, page in enumerate(client.paginate(GetDocumentReport(limit=10))):
        document_report.extend(page)
        if i > num_pages_to_check:
            break
    assert document_report is not None
    assert len(document_report) > 0


def test_all_submissions(indico):
    """
    NOTE: This test is expected to fail unless user has ALL_SUBMISSION_LOGS scope
    """
    client = IndicoClient()
    page = next(client.paginate(GetDocumentReport(limit=1000, all_submissions=True)))
    assert page is not None



from pathlib import Path
import time
from typing import List

import pytest
from indico.client import IndicoClient
from indico.queries import (
    JobStatus,
    RetrieveStorageObject,
    UpdateWorkflowSettings,
    WorkflowSubmission,
    WaitForSubmissions,
    SubmitReview,
)

from indico.types.workflow_metrics import WorkflowMetrics, WorkflowMetricsOptions
from indico.queries.workflow_metrics import GetWorkflowMetrics
from datetime import datetime
from ..data.datasets import *  # noqa
from ..data.datasets import PUBLIC_URL
import time


@pytest.fixture
def workflow(indico, org_annotate_dataset, org_annotate_workflow, org_annotate_model_group):
    client = IndicoClient()

    wf = client.call(
        UpdateWorkflowSettings(org_annotate_workflow.id, enable_review=True, enable_auto_review=True)
    )
    assert wf.review_enabled and wf.auto_review_enabled

    _file = str(Path(__file__).parents[1]) + "/data/org-sample.pdf"

    sub_ids = client.call(WorkflowSubmission(workflow_id=wf.id, files=[_file]))
    subs = client.call(WaitForSubmissions(sub_ids, timeout=900))
    sub = subs[0]
    assert sub.status == "PENDING_AUTO_REVIEW"
    raw_result = client.call(RetrieveStorageObject(sub.result_file))
    changes = raw_result["results"]["document"]["results"]
    for model, preds in changes.items():
        if isinstance(preds, dict):
            preds["accepted"] = True
        elif isinstance(preds, list):
            for pred in preds:
                pred["accepted"] = True
    job = client.call(SubmitReview(sub.id, changes=changes, force_complete=True))
    job = client.call(JobStatus(job.id))
    submission = client.call(WaitForSubmissions([sub.id]))[0]
    assert submission.status == "COMPLETE"

    # metrics in the cluster are updated every 5 minutes
    time.sleep(300)
    return wf


def test_fetch_metrics(indico, workflow):
    client = IndicoClient()
    workflow_metric: List[WorkflowMetrics] = client.call(
        GetWorkflowMetrics(
            options=[WorkflowMetricsOptions.SUBMISSIONS],
            start_date=datetime.now(),
            end_date=datetime.now(),
            workflow_ids=[workflow.id],
        )
    )
    assert workflow_metric is not None
    assert workflow_metric[0].submissions is not None
    assert workflow_metric[0].submissions.aggregate.submitted > 0


def test_fetch_metrics_queue(indico, workflow):
    client = IndicoClient()
    workflow_metric: List[WorkflowMetrics] = client.call(
        GetWorkflowMetrics(
            options=[WorkflowMetricsOptions.REVIEW],
            start_date=datetime.now(),
            end_date=datetime.now(),
            workflow_ids=[workflow.id],
        )
    )
    assert workflow_metric is not None
    assert workflow_metric[0].queues is not None
    assert len(workflow_metric[0].queues.daily_cumulative) > 0



from typing import Generator, Optional, Union

import requests

from indico.client.request import GraphQLRequest, RequestChain
from indico.errors import IndicoInputError, IndicoRequestError
from indico.queries.jobs import JobStatus
from indico.types.jobs import Job


class _UploadSMExport(GraphQLRequest):
    query = """
        query exportUpload {
            exportUpload {
                signedUrl
                storageUri
            }
        }
    """

    def __init__(self, file_path: str):
        self.file_path = file_path
        super().__init__(self.query)

    def process_response(self, response) -> str:
        resp = super().process_response(response)["exportUpload"]
        signed_url = resp["signedUrl"]
        storage_uri = resp["storageUri"]

        with open(self.file_path, "rb") as file:
            file_content = file.read()

        headers = {"Content-Type": "application/zip"}
        response = requests.put(signed_url, data=file_content, headers=headers)

        if response.status_code != 200:
            raise IndicoRequestError(
                f"Failed to upload static model export: {response.text}"
            )
        return storage_uri


class ProcessStaticModelExport(GraphQLRequest):
    """
    Process a static model export.

    Available on 6.14+ only.

    Args:
        storage_uri(str): the storage uri of the static model export that was uploaded to Indico.
        workflow_id(int): the id of the workflow that the static model export is being added to.

    Returns:
        Job: the id of the job that is processing the static model export.
    """

    query = """
        mutation processStaticModelExport($storageUri: String!, $workflowId: Int!) {
            processStaticModelExport(storageUri: $storageUri, workflowId: $workflowId) {
                jobId
            }
        }
    """

    def __init__(
        self,
        storage_uri: str,
        workflow_id: int,
    ):
        self.storage_uri = storage_uri
        self.workflow_id = workflow_id
        super().__init__(
            self.query,
            variables={
                "storageUri": self.storage_uri,
                "workflowId": self.workflow_id,
            },
        )

    def process_response(self, response) -> Job:
        job_id = super().process_response(response)["processStaticModelExport"]["jobId"]
        return Job(id=job_id)


class UploadStaticModelExport(RequestChain):
    """
    Upload a static model export to Indico.

    Available on 6.14+ only.

    Args:
        file_path(str): path to the export zip file to upload to Indico.

    Options:
        auto_process(bool): if True, the static model export will be processed after it is uploaded.
        workflow_id(int): the id of the workflow that the static model export is being added to. Required if `auto_process` is True.

    Returns:
        str: the storage uri of the static model export.
    """

    def __init__(
        self,
        file_path: str,
        auto_process: bool = False,
        workflow_id: Optional[int] = None,
    ):
        self.file_path = file_path
        self.auto_process = auto_process
        if auto_process and not workflow_id:
            raise IndicoInputError(
                "Must provide `workflow_id` if `auto_process` is True."
            )

        self.workflow_id = workflow_id

    def requests(self) -> Generator[Union[str, Job], None, None]:
        if self.auto_process:
            yield _UploadSMExport(self.file_path)
            yield ProcessStaticModelExport(
                storage_uri=self.previous, workflow_id=self.workflow_id
            )
            yield JobStatus(self.previous.id)
            if self.previous.status == "FAILURE":
                raise IndicoRequestError(
                    code="FAILURE",
                    error=f"Failed to process static model export: {self.previous.result['message']}",
                )
        else:
            yield _UploadSMExport(self.file_path)



import asyncio
from random import randint
import time
import typing as t
from functools import wraps



def retry(
    ExceptionTypes: t.Type[Exception], tries: int = 3, delay: int = 1, backoff: int = 2
) -> t.Callable:
    """
    Retry with exponential backoff

    Original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry
    """

    def retry_decorator(f: t.Callable) -> t.Any:
        @wraps(f)
        def retry_fn(*args: t.Any, **kwargs: t.Any) -> t.Any:
            n_tries, n_delay = tries, delay
            while n_tries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionTypes:
                    time.sleep(n_delay)
                    n_tries -= 1
                    n_delay *= backoff
            return f(*args, **kwargs)

        return retry_fn

    return retry_decorator

def aioretry(
    ExceptionTypes: t.Type[Exception],
    tries: int = 3,
    delay: t.Union[int, t.Tuple[int, int]] = 1,
    backoff: int = 2,
    condition: t.Optional[t.Callable[[Exception], bool]] = None,
) -> t.Callable:
    """
    Retry with exponential backoff

    Original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry
    Options:
        condition: Callable to evaluate if an exception of a given type
            is retryable for additional handling
        delay: an initial time to wait (seconds). If a tuple, choose a random number
            in that range to start. This can helps prevent retries at the exact
            same time across multiple concurrent function calls
    """

    def retry_decorator(f: t.Callable) -> t.Callable:
        @wraps(f)
        async def retry_fn(*args: t.Any, **kwargs: t.Any) -> t.Any:
            n_tries = tries
            if isinstance(delay, tuple):
                # pick a random number to sleep
                n_delay = randint(*delay)
            else:
                n_delay = delay
            while True:
                try:
                    return await f(*args, **kwargs)
                except ExceptionTypes as e:
                    if condition and not condition(e):
                        raise
                    await asyncio.sleep(n_delay)
                    n_tries -= 1
                    n_delay *= backoff
                    if n_tries <= 0:
                        raise

        return retry_fn

    return retry_decorator



import asyncio
import http.cookiejar
import logging
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path
from typing import Optional, Union

import aiohttp
import requests

from indico.client.request import HTTPRequest
from indico.config import IndicoConfig
from indico.errors import (
    IndicoAuthenticationFailed,
    IndicoHibernationError,
    IndicoRequestError,
)
from indico.http.serialization import aio_deserialize, deserialize

from .retry import aioretry

logger = logging.getLogger(__file__)


class CookiePolicyOverride(http.cookiejar.DefaultCookiePolicy):
    def set_ok(self, cookie, request):
        return True

    def return_ok(self, cookie, request):
        return True

    def path_return_ok(self, path, request):
        return True

    def domain_return_ok(self, domain, request):
        return True


class HTTPClient:
    def __init__(self, config: IndicoConfig = None):
        self.config = config
        self.base_url = f"{self.config.protocol}://{self.config.host}"

        self.request_session = requests.Session()
        if config and isinstance(config.requests_params, dict):
            for param in config.requests_params.keys():
                setattr(self.request_session, param, config.requests_params[param])
        self.request_session.cookies.set_policy(CookiePolicyOverride())
        self.get_short_lived_access_token()

    def post(self, *args, json: Union[dict, list] = None, **kwargs):
        return self._make_request("post", *args, json=json, **kwargs)

    def get(self, *args, params: dict = None, **kwargs):
        return self._make_request("post", *args, params=params, **kwargs)

    def get_short_lived_access_token(self):
        # If the cookie here is already due to _disable_cookie_domain set and we try to pop it later
        # it will error out because we have two cookies with the same name. We just remove the old one
        # here as we are about to refresh it.
        if "auth_token" in self.request_session.cookies:
            self.request_session.cookies.pop("auth_token")

        r = self.post(
            "/auth/users/refresh_token",
            headers={"Authorization": f"Bearer {self.config.api_token}"},
            _refreshed=True,
        )

        # Disable cookie domain in cases where the domain won't match due to using short name domains
        if self.config._disable_cookie_domain:
            value = self.request_session.cookies.get("auth_token", None)
            if not value:
                raise IndicoAuthenticationFailed()
            self.request_session.cookies.pop("auth_token")
            self.request_session.cookies.set_cookie(
                requests.cookies.create_cookie(name="auth_token", value=value)
            )

        return r

    def execute_request(self, request: HTTPRequest):
        return request.process_response(
            self._make_request(
                method=request.method.value.lower(), path=request.path, **request.kwargs
            )
        )

    @contextmanager
    def _handle_files(self, req_kwargs):
        streams = None
        # deepcopying buffers is not supported
        # so, remove "streams" before the deepcopy.
        if "streams" in req_kwargs:
            if req_kwargs["streams"] is not None:
                streams = req_kwargs["streams"].copy()
            del req_kwargs["streams"]

        new_kwargs = deepcopy(req_kwargs)

        files = []
        file_arg = {}
        dup_counts = {}
        if "files" in new_kwargs and new_kwargs["files"] is not None:
            for filepath in new_kwargs["files"]:
                path = Path(filepath)
                fd = path.open("rb")
                files.append(fd)
                # follow the convention of adding (n) after a duplicate filename
                if path.stem in dup_counts:
                    file_arg[path.stem + f"({dup_counts[path.stem]})"] = fd
                    dup_counts[path.stem] += 1
                else:
                    file_arg[path.stem] = fd
                    dup_counts[path.stem] = 1

        if streams is not None and len(streams) > 0:
            for filename in streams:
                # similar operation as above.
                stream = streams[filename]
                files.append(stream)
                if filename in dup_counts:
                    file_arg[filename + f"({dup_counts[filename]})"] = stream
                    dup_counts[filename] += 1
                else:
                    file_arg[filename] = stream
                    dup_counts[filename] = 1

        new_kwargs["files"] = file_arg

        yield new_kwargs

        if files:
            [f.close() for f in files]

    def _make_request(
        self,
        method: str,
        path: str,
        headers: dict = None,
        _refreshed=False,
        **request_kwargs,
    ):
        logger.debug(
            f"[{method}] {path}\n\t Headers: {headers}\n\tRequest Args:{request_kwargs}"
        )
        with self._handle_files(request_kwargs) as new_kwargs:
            response = getattr(self.request_session, method)(
                f"{self.base_url}{path}",
                headers=headers,
                stream=True,
                verify=False
                if not self.config.verify_ssl or not self.request_session.verify
                else True,
                **new_kwargs,
            )

        json: bool = ".json" in Path(path).suffixes
        decompress: bool = Path(path).suffix == ".gz"

        # If auth expired refresh
        if response.status_code == 401 and not _refreshed:
            self.get_short_lived_access_token()
            return self._make_request(
                method, path, headers, _refreshed=True, **request_kwargs
            )
        elif response.status_code == 401 and _refreshed:
            raise IndicoAuthenticationFailed()

        if response.status_code == 503 and "Retry-After" in response.headers:
            raise IndicoHibernationError(after=response.headers.get("Retry-After"))

        if response.status_code >= 500:
            raise IndicoRequestError(
                code=response.status_code,
                error=response.reason,
                extras=repr(response.content),
            )

        content = deserialize(response, force_json=json, force_decompress=decompress)

        if response.status_code >= 400:
            if isinstance(content, dict):
                error = (
                    f"{content.pop('error_type', 'Unknown Error')}, "
                    f"{content.pop('message', '')}"
                )
                extras = content
            else:
                error = content
                extras = None

            raise IndicoRequestError(
                error=error, code=response.status_code, extras=extras
            )
        return content


class AIOHTTPClient(HTTPClient):
    """
    Beta client with a minimal set of features that can execute
    requests using the aiohttp library
    """

    def __init__(self, config: Optional[IndicoConfig] = None):
        """
        Config options specific to aiohttp
        unsafe - allows interacting with IP urls
        """
        self.config = config or IndicoConfig()
        self.base_url = f"{self.config.protocol}://{self.config.host}"

        self.request_session = aiohttp.ClientSession()
        if config and isinstance(config.requests_params, dict):
            for param in config.requests_params.keys():
                setattr(self.request_session, param, config.requests_params[param])

    async def post(self, *args, json: Union[dict, list] = None, **kwargs):
        return await self._make_request("post", *args, json=json, **kwargs)

    async def get(self, *args, params: dict = None, **kwargs):
        return await self._make_request("post", *args, params=params, **kwargs)

    async def get_short_lived_access_token(self):
        r = await self.post(
            "/auth/users/refresh_token",
            headers={"Authorization": f"Bearer {self.config.api_token}"},
            _refreshed=True,
        )
        return r

    async def execute_request(self, request: HTTPRequest):
        return request.process_response(
            await self._make_request(
                method=request.method.value.lower(), path=request.path, **request.kwargs
            )
        )

    @contextmanager
    def _handle_files(self, req_kwargs):
        files = []
        file_args = []
        dup_counts = {}
        for filepath in req_kwargs.pop("files", []) or []:
            data = aiohttp.FormData()
            path = Path(filepath)
            fd = path.open("rb")
            files.append(fd)
            # follow the convention of adding (n) after a duplicate filename
            _add_suffix = f".{path.suffix}" if path.suffix else ""
            if path.stem in dup_counts:
                data.add_field(
                    "file",
                    fd,
                    filename=path.stem + f"({dup_counts[path.stem]})" + _add_suffix,
                )
                dup_counts[path.stem] += 1
            else:
                data.add_field("file", fd, filename=path.name)
                dup_counts[path.stem] = 1
            file_args.append(data)

        for filename, stream in (req_kwargs.pop("streams", {}) or {}).items():
            # similar operation as above.
            files.append(stream)
            data = aiohttp.FormData()
            if filename in dup_counts:
                data.add_field(
                    "file",
                    stream,
                    filename=filename + f"({dup_counts[filename]})" + _add_suffix,
                )
                dup_counts[filename] += 1
            else:
                data.add_field("file", stream, filename=filename)
                dup_counts[filename] = 1
            file_args.append(data)

        yield file_args

        if files:
            [f.close() for f in files]

    @aioretry((aiohttp.ClientConnectionError, aiohttp.ServerDisconnectedError))
    async def _make_request(
        self,
        method: str,
        path: str,
        headers: dict = None,
        _refreshed=False,
        **request_kwargs,
    ):
        logger.debug(
            f"[{method}] {path}\n\t Headers: {headers}\n\tRequest Args:{request_kwargs}"
        )
        json: bool = ".json" in Path(path).suffixes
        decompress: bool = Path(path).suffix == ".gz"

        with self._handle_files(request_kwargs) as file_args:
            if file_args:
                resps = await asyncio.gather(
                    *(
                        self._make_request(
                            method, path, headers, **request_kwargs, data=data
                        )
                        for data in file_args
                    )
                )
                return [resp for resp_set in resps for resp in resp_set]
            async with getattr(self.request_session, method)(
                f"{self.base_url}{path}",
                headers=headers,
                verify_ssl=self.config.verify_ssl,
                **request_kwargs,
            ) as response:
                # If auth expired refresh
                if response.status == 401 and not _refreshed:
                    await self.get_short_lived_access_token()
                    return await self._make_request(
                        method, path, headers, _refreshed=True, **request_kwargs
                    )
                elif response.status == 401 and _refreshed:
                    raise IndicoAuthenticationFailed()

                if response.status == 503 and "Retry-After" in response.headers:
                    raise IndicoHibernationError(
                        after=response.headers.get("Retry-After")
                    )

                if response.status >= 500:
                    raise IndicoRequestError(
                        code=response.status_code,
                        error=response.reason,
                        extras=repr(response.content),
                    )

                content = await aio_deserialize(
                    response, force_json=json, force_decompress=decompress
                )

                if response.status >= 400:
                    if isinstance(content, dict):
                        error = (
                            f"{content.pop('error_type', 'Unknown Error')}, "
                            f"{content.pop('message', '')}"
                        )
                        extras = content
                    else:
                        error = content
                        extras = None

                    raise IndicoRequestError(
                        error=error, code=response.status_code, extras=extras
                    )
                return content



""""""
from typing import List, Union

from indico import PagedRequest
from indico.filters import DocumentReportFilter
from indico.types import BaseType
from indico.types.document_report import DocumentReport


class _DocumentReportList(BaseType):
    submissions: List[DocumentReport]
    pass


class GetDocumentReport(PagedRequest):
    """
    Query to generate a Document Report, otherwise known as a log of past submissions.
    """

    query = """
       query SubmissionsLog($filters: SubmissionLogFilter, $limit: Int, $after: Int, $allSubmissions: Boolean){
          submissionsLog(filters: $filters, limit: $limit, after: $after, allSubmissions: $allSubmissions){
            submissions{
              datasetId
              workflowId
              status
              createdAt
              createdBy
              updatedAt
              updatedBy
              completedAt
              errors
              retrieved
              submissionId
              filesDeleted
              inputFiles{
                id
                filename
                filepath
                submissionId
                fileSize
                numPages
              }
            }
            pageInfo{
              startCursor
              endCursor
              hasNextPage
              aggregateCount
            }
          }
        }
        """

    def __init__(
        self, filters: Union[dict, DocumentReportFilter] = None, limit: int = None, all_submissions = False
    ):
        variables = {"filters": filters, "limit": limit, "allSubmissions": all_submissions}
        super().__init__(self.query, variables=variables)

    def process_response(self, response):
        return _DocumentReportList(
            **super().process_response(response)["submissionsLog"]
        ).submissions



import re
from typing import Union
import time
from indico.errors import IndicoTimeoutError

_cc_to_snake_re = re.compile(r"(?<!^)(?=[A-Z])")
_snake_to_cc_re = re.compile(r"(.*?)_([a-zA-Z])")


def cc_to_snake(string: str):
    return re.sub(_cc_to_snake_re, "_", string).lower()


def _camel(match):
    return match.group(1) + match.group(2).upper()


def snake_to_cc(string: str):
    return re.sub(_snake_to_cc_re, _camel, string, 0)



class Timer:
    def __init__(self, timeout: Union[int, float]):
        self.timeout = timeout
        self.start = time.time()
        self.elapsed = 0

    def check(self):
        self.elapsed = time.time() - self.start
        if self.timeout < self.elapsed:
            raise IndicoTimeoutError(self.elapsed)


import datetime

from indico.types import BaseType


class OutputFile(BaseType):
    """
    An Output File in the Indico Platform.

    Attributes:
        id (int): The Output file id
        filepath (str): URL of the output datafile within the Indico Platform.
        submission_id (int): The parent Submission id
        component_id (int): The id of the corresponding component the output is linked to
        created_at (datetime): The date the output file was created
    """

    id: int
    filepath: str
    submission_id: int
    component_id: int
    created_at: datetime



import datetime
from typing import List, Union, Dict

from indico.client.request import (
    GraphQLRequest,
    PagedRequest,
)
from indico.filters import UserMetricsFilter
from indico.types import BaseType
from indico.types.user_metrics import UserSummary, UserSnapshot, UserChangelog, UserChangelogReport


class _PagedUserSnapshots(BaseType):
    """
    Class to hold paged snapshot data to make parsing easier.
    """
    results: List[UserSnapshot]


class _PagedUserChangelog(BaseType):
    """
    Class to hold paged snapshot data to make parsing easier.
    """
    results: List[UserChangelog]


class GetUserSummary(GraphQLRequest):
    """
    Requests summary-level information of users in the app on a specific date.

    Args:
        date (datetime): specific day to summarize.

    """
    query = """
query GetUserSummary($date: Date){
  userSummary(date: $date){
    users{
      enabled
      disabled
    }
    appRoles {
      role
      count
    }
  }
}
    """

    def __init__(self, date=None):
        if date is not None:
            super().__init__(self.query, variables={"date": date.strftime('%Y-%m-%d')})
        else:
            super().__init__(self.query)

    def process_response(self, response) -> UserSummary:
        return UserSummary(**super().process_response(response)["userSummary"])


class GetUserSnapshots(PagedRequest):
    """

    Requests paged detailed information about app users on a specific date.

    Args:
        filters (UserMetricsFilter): filter the query based on UserMetricsFilter criteria.
        date (datetime): specific day to query.
        limit (int): limit how many come back per query or per page.
    """
    query = """
    query GetUserSnapshot($date: Date, $filters: UserReportFilter, $after: Int, $limit: Int){
  userSnapshot(date: $date, filters: $filters, after: $after, limit: $limit){
    results{
      id
      name
      email
      createdAt
      enabled
      roles
      datasets{
        datasetId
        role
      }
      
    }
    pageInfo{
      startCursor
      endCursor
      hasNextPage
      aggregateCount
    }
  }
}
    """

    def __init__(self, *, date: datetime, filters: Union[Dict, UserMetricsFilter] = None, limit: int = None):
        variables = {
            "date": date.strftime('%Y-%m-%d') if date is not None else None,
            "filters": filters,
            "limit": limit
        }
        super().__init__(self.query, variables=variables)

    def process_response(self, response) -> List[UserSnapshot]:
        return _PagedUserSnapshots(**super().process_response(response)["userSnapshot"]).results


class GetUserChangelog(PagedRequest):
    """

        Gets paged detailed information about app users.

        Args:
            filters (UserSnapshotFilter): filter the query based on UserMetricsFilter criteria.
            start_date (datetime): specific start date for query.
            end_date (datetime): specific end date for query.
            limit (int): limit how many come back per query or per page.
        """
    query = """
        query GetUserChangelog($sdate: Date, $edate: Date, $filters: UserReportFilter, $after: Int, $limit: Int){
        userChangelog(startDate: $sdate, endDate: $edate, filters: $filters, after:$after, limit:$limit){
            results {
            id
            date
            userEmail
            updaterEmail
            previousDatasets {
                datasetId
                role
            }
            changesMade
        }
        pageInfo{
            aggregateCount
            hasNextPage
            endCursor
        }

        }
    }
        """

    def __init__(self, *, start_date: datetime, end_date: datetime, filters: Union[Dict, UserMetricsFilter] = None,
                 limit: int = None):
        variables = {
            "sdate": start_date.strftime('%Y-%m-%d') if start_date is not None else None,
            "edate": end_date.strftime('%Y-%m-%d') if end_date is not None else None,
            "filters": filters,
            "limit": limit
        }
        super().__init__(self.query, variables=variables)

    def process_response(self, response) -> List[UserSnapshot]:
        return _PagedUserChangelog(**super().process_response(response)["userChangelog"]).results


class GenerateChangelogReport(GraphQLRequest):
    """

        Creates a job to generate a report of detailed information about app users
        
        Args:
            filters (UserSnapshotFilter): filter the query based on UserMetricsFilter criteria.
            start_date (datetime): specific start date for query.
            end_date (datetime): specific end date for query.
            report_format (str): specific format of the report, JSON or CSV.

        """
    query = """
        mutation GenerateChangeReport ($sdate: Date, $edate: Date, $filters: UserReportFilter){
    userChangelogReport(
        startDate: $sdate,
        endDate: $edate,
        filters: $filters,
        reportFormat: JSON
    ){
        jobId
    }
}
        """

    def __init__(self, *, start_date: datetime, end_date: datetime, filters: Union[Dict, UserMetricsFilter] = None,
                 report_format: str = "csv"):
        variables = {
            "sdate": start_date.strftime('%Y-%m-%d') if start_date is not None else None,
            "edate": end_date.strftime('%Y-%m-%d') if end_date is not None else None,
            "filters": filters,
            "format": report_format
        }
        super().__init__(self.query, variables=variables)

    def process_response(self, response) -> List[UserSnapshot]:
        return UserChangelogReport(**super().process_response(response)["userChangelogReport"])



from typing import Any, Dict, List, Optional

import jsons

from indico import GraphQLRequest, RequestChain
from indico.errors import IndicoInputError
from indico.queries.model_import import (
    UploadStaticModelExport,
)
from indico.types import (
    LinkedLabelGroup,
    NewLabelsetArguments,
    NewQuestionnaireArguments,
    Workflow,
)


class _AddWorkflowComponent(GraphQLRequest):
    query = """
        mutation addWorkflowComponent($afterComponentId: Int, $afterComponentLinkId: Int, $component: JSONString!, $workflowId: Int!, $blueprintId: Int) {
        addWorkflowComponent(
            afterComponentId: $afterComponentId
            component: $component
            workflowId: $workflowId
            afterComponentLinkId: $afterComponentLinkId
            blueprintId: $blueprintId
        ) {
            workflow {
            id
            components {
                id
                componentType
                reviewable
                ... on ContentLengthComponent {
                minimum
                maximum
                }
                filteredClasses
                ... on ModelGroupComponent {
                taskType
                modelType
                modelGroup {
                    status
                    id
                    classNames
                    name
                    taskType
                    questionnaireId
                    selectedModel {
                    id
                    }
                }
                }
            }
            componentLinks {
                id
                headComponentId
                tailComponentId
                filters {
                classes
                }
            }
            }
        }
        }
    """

    def __init__(
        self,
        after_component_id: int,
        after_component_link: int,
        workflow_id: int,
        component: dict,
        blueprint_id: Optional[int] = None,
    ):
        super().__init__(
            self.query,
            variables={
                "afterComponentId": after_component_id,
                "afterComponentLinkId": after_component_link,
                "workflowId": workflow_id,
                "component": jsons.dumps(component),
                "blueprintId": blueprint_id,
            },
        )

    def process_response(self, response) -> Workflow:
        return Workflow(
            **super().process_response(response)["addWorkflowComponent"]["workflow"]
        )


class AddLinkedLabelComponent(RequestChain):
    """
    Adds a linked label transformer that groups together labels

    Args:
        after_component_id(int): the component this component follows.
        after_component_link_id(int): the component link this component follows.
        workflow_id(int): the workflow id.
        labelset_id(int): the labelset to source classes from.
        model_group_id(int): the model group to source classes from.
        groups (List[LinkedLabelGroup]): configuration for how to group labels.
    """

    def __init__(
        self,
        after_component_id: int,
        workflow_id: int,
        labelset_id: int,
        model_group_id: int,
        groups: List[LinkedLabelGroup],
        after_component_link_id: int = None,
    ):
        self.workflow_id = workflow_id
        self.after_component_id = after_component_id
        self.after_component_link_id = after_component_link_id
        self.component = {
            "component_type": "link_label",
            "config": {
                "labelset_id": labelset_id,
                "model_group_id": model_group_id,
                "groups": [self.__groups_to_json(group) for group in groups],
            },
        }

    def __groups_to_json(self, group: LinkedLabelGroup):
        return {
            "name": group.name,
            "strategy": group.strategy.name.lower(),
            "class_names": group.class_ids,
            "strategy_settings": group.strategy_settings,
        }

    def requests(self):
        yield _AddWorkflowComponent(
            after_component_id=self.after_component_id,
            after_component_link=self.after_component_link_id,
            workflow_id=self.workflow_id,
            component=self.component,
        )


class AddContentLengthFilterComponent(RequestChain):
    """
    Adds a content length filter.

    Args:
        workflow_id(int): the workflow id.
        after_component_id(int): the component this component follows.
        minimum(int): minimum length of content to accept. Defaults to None.
        maximum(int): maximum length of content to accept. Defaults to None.
    """

    def __init__(
        self,
        workflow_id: int,
        after_component_id: int,
        after_component_link_id: int = None,
        minimum: int = None,
        maximum: int = None,
    ):
        self.workflow_id = workflow_id
        self.after_component_id = after_component_id
        self.after_component_link_id = after_component_link_id
        self.minimum = minimum
        self.maximum = maximum
        self.component = {
            "component_type": "content_length",
            "config": {"minimum": minimum, "maximum": maximum},
        }

    def requests(self):
        yield _AddWorkflowComponent(
            after_component_id=self.after_component_id,
            after_component_link=self.after_component_link_id,
            workflow_id=self.workflow_id,
            component=self.component,
        )


class AddLinkClassificationComponent(RequestChain):
    """
    Adds a link classification model component with filtered classes.

    Args:
        workflow_id(int): the workflow id.
        after_component_id(int): the component this component follows.
        model_group_id(int): the model group to source classes from.
        filtered_classes(List[List[str]]): sets of classes to filter, i.e., [["A"], ["A","B"]]
        labels(str): decides if to use "actual" or "predicted" labels.
    """

    def __init__(
        self,
        workflow_id: int,
        after_component_id: int,
        model_group_id: int,
        filtered_classes: List[List[str]],
        labels: str = None,
        after_component_link_id: int = None,
    ):
        self.workflow_id = workflow_id
        self.after_component_id = after_component_id
        self.after_component_link_id = after_component_link_id
        self.component = {
            "component_type": "link_classification_model",
            "config": {
                "model_group_id": model_group_id,
                "filtered_classes": filtered_classes,
                "labels": labels,
            },
        }

    def requests(self):
        yield _AddWorkflowComponent(
            after_component_id=self.after_component_id,
            after_component_link=self.after_component_link_id,
            workflow_id=self.workflow_id,
            component=self.component,
        )


class AddModelGroupComponent(GraphQLRequest):
    """
    Adds a new model group to a workflow, optionally with a customized questionnaire.
    Available on 5.0+ only.
    Returns workflow with updated component list, which will contain the added Model Group.

    Args:
        workflow_id(int): the id of the workflow to add the component to.
        dataset_id(int): the id of the dataset of this workflow/model group.
        name(str): name of the model group.
        source_column_id(str): source column identifier from dataset.
        after_component_id(str): the id of the previous component, or step, that should precede this step.
            Typically, this prior component would be something like "INPUT_OCR_EXTRACTION" or "INPUT_IMAGE" and can
            be found by using "component_by_type()" on the parent workflow.
        labelset_column_id(int): the labelset column to copy from.
        new_labelset_args(NewLabelsetArguments): if needed, new labelset to add.
            Only use if not using labelset_column_id.
        new_questionnaire_args(NewQuestionnaireArguments): Customize the questionnaire associated with this model group.

    """

    query = """
        mutation addModelGroup($workflowId: Int!, $name: String!, $datasetId: Int!, $sourceColumnId: Int!, $afterComponentId: Int, $labelsetColumnId: Int, $afterLinkId: Int, $newLabelsetArgs: NewLabelsetInput, $questionnaireArgs: QuestionnaireInput, $modelTrainingOptions: JSONString, $modelType: ModelType) {
        addModelGroupComponent(
            workflowId: $workflowId
            name: $name
            datasetId: $datasetId
            sourceColumnId: $sourceColumnId
            afterComponentId: $afterComponentId
            afterLinkId: $afterLinkId
            labelsetColumnId: $labelsetColumnId
            modelTrainingOptions: $modelTrainingOptions
            newLabelsetArgs: $newLabelsetArgs
            questionnaireArgs: $questionnaireArgs
            modelType: $modelType
        ) {
            workflow {
            id
            components {
                id
                componentType
                reviewable
                filteredClasses
                ... on ContentLengthComponent {
                minimum
                maximum
                }
                ... on ModelGroupComponent {
                taskType
                modelType
                modelGroup {
                    status
                    id
                    name
                    taskType
                    classNames
                    questionnaireId
                    selectedModel {
                    id
                    }
                }
                }
            }
            componentLinks {
                id
                headComponentId
                tailComponentId
                filters {
                classes
                }
            }
            }
        }
        }
    """

    def __init__(
        self,
        workflow_id: int,
        dataset_id: int,
        name: str,
        source_column_id: int,
        after_component_id: int = None,
        after_link_id: int = None,
        labelset_column_id: int = None,
        new_labelset_args: NewLabelsetArguments = None,
        new_questionnaire_args: NewQuestionnaireArguments = None,
        model_training_options: str = None,
        model_type: str = None,
    ):
        if labelset_column_id is not None and new_labelset_args is not None:
            raise IndicoInputError(
                "Cannot define both labelset_column_id and new_labelset_args, must be one "
                "or the other."
            )
        if labelset_column_id is None and new_labelset_args is None:
            raise IndicoInputError(
                "Must define one of either labelset_column_id or new_labelset_args."
            )

        if model_training_options:
            model_training_options = jsons.dumps(model_training_options)

        super().__init__(
            self.query,
            variables={
                "workflowId": workflow_id,
                "name": name,
                "datasetId": dataset_id,
                "sourceColumnId": source_column_id,
                "labelsetColumnId": labelset_column_id,
                "afterComponentId": after_component_id,
                "afterLinkId": after_link_id,
                "modelTrainingOptions": model_training_options,
                "modelType": model_type,
                "newLabelsetArgs": (
                    self.__labelset_to_json(new_labelset_args)
                    if new_labelset_args is not None
                    else None
                ),
                "questionnaireArgs": (
                    self.__questionnaire_to_json(new_questionnaire_args)
                    if new_questionnaire_args is not None
                    else None
                ),
            },
        )

    def __labelset_to_json(self, labelset: NewLabelsetArguments):
        return {
            "name": labelset.name,
            "numLabelersRequired": labelset.num_labelers_required,
            "datacolumnId": labelset.datacolumn_id,
            "taskType": labelset.task_type.name,
            "targetNames": labelset.target_names,
        }

    def __questionnaire_to_json(self, questionnaire: NewQuestionnaireArguments):
        return {
            "instructions": questionnaire.instructions,
            "forceTextMode": questionnaire.force_text_mode,
            "showPredictions": questionnaire.show_predictions,
            "users": questionnaire.users,
        }

    def process_response(self, response) -> Workflow:
        return Workflow(
            **super().process_response(response)["addModelGroupComponent"]["workflow"]
        )


class DeleteWorkflowComponent(GraphQLRequest):
    """
    Deletes a component from a workflow. If the component has an associated model, the model is deleted as well.
    Available on 5.3+ only.
    Returns workflow with updated list of components and links

    Args:
        workflow_id(int): the id of the workflow to delete the component from.
        component_id(int): the id of the component to delete.
    """

    query = """
        mutation deleteWorkflowComponent(
            $workflowId: Int!,
            $componentId: Int!
        ){
            deleteWorkflowComponent(
                workflowId: $workflowId,
                componentId: $componentId
            ){
                workflow {
                    id
                    components {
                        id
                        componentType
                        reviewable
                        filteredClasses
                        ... on ContentLengthComponent {
                            minimum
                            maximum
                        }
                        ... on ModelGroupComponent {
                            taskType
                            modelType
                            modelGroup {
                                status
                                id
                                name
                                taskType
                                questionnaireId
                                selectedModel{
                                    id
                                }
                            }
                        }

                    }
                    componentLinks {
                        id
                        headComponentId
                        tailComponentId
                        filters {
                            classes
                        }
                    }
                }
            }
        }
    """

    def __init__(self, workflow_id: int, component_id: int):
        super().__init__(
            self.query,
            variables={"workflowId": workflow_id, "componentId": component_id},
        )

    def process_response(self, response) -> Workflow:
        return Workflow(
            **super().process_response(response)["deleteWorkflowComponent"]["workflow"]
        )


class AddStaticModelComponent(RequestChain):
    """
    Add a static model component to a workflow.

    Available on 6.14+ only.

    Args:
        `workflow_id(int)`: the id of the workflow to add the component to.
        `after_component_id(int)`: the id of the component to add this component after. Should be after the input ocr extraction component.
        `after_component_link_id(int)`: the component link to add this component after.
        `static_component_config(dict[str, Any])`: the configuration for the static model component. this would the dictionary returned from the job upon completion.
        `auto_process(bool)`: if True, the static model export will be automatically processed after it is uploaded.
        `export_file(str)`: the path to the static model export file.
    """

    previous = None

    def __init__(
        self,
        workflow_id: int,
        after_component_id: Optional[int] = None,
        after_component_link_id: Optional[int] = None,
        static_component_config: Optional[Dict[str, Any]] = None,
        component_name: Optional[str] = None,
        auto_process: bool = False,
        export_file: Optional[str] = None,
    ):
        if not export_file and auto_process:
            raise IndicoInputError("Must provide export_file if auto_process is True.")

        if not auto_process and not static_component_config:
            raise IndicoInputError(
                "Must provide static_component_config if auto_process is False."
            )

        if not after_component_id and not after_component_link_id:
            raise IndicoInputError(
                "Must provide either `after_component_id` or `after_component_link_id`."
            )

        self.workflow_id = workflow_id
        self.after_component_id = after_component_id
        self.after_component_link_id = after_component_link_id
        self.component = {
            "component_type": "static_model",
            "config": {
                "export_meta": static_component_config,
            },
        }

        if component_name:
            self.component.update({"name": component_name})

        self.auto_process = auto_process
        self.export_file = export_file

    def requests(self):
        if self.auto_process:
            yield UploadStaticModelExport(
                auto_process=True,
                file_path=self.export_file,
                workflow_id=self.workflow_id,
            )
            self.component.update(
                {
                    "config": {
                        "export_meta": self.previous.result,
                    }
                }
            )

        yield _AddWorkflowComponent(
            after_component_id=self.after_component_id,
            after_component_link=self.after_component_link_id,
            workflow_id=self.workflow_id,
            component=self.component,
        )



from typing import List, Optional, Union

from indico.client.request import Delay, GraphQLRequest, RequestChain
from indico.types.model_export import ModelExport


class _CreateModelExport(GraphQLRequest):
    query = """
        mutation ($modelId: Int!) {
            createModelExport(
                modelId: $modelId
            ) {
                id
                name
                status
                modelId
            }
        }
    """

    def __init__(self, model_id: int):
        self.model_id = model_id
        super().__init__(self.query, variables={"modelId": model_id})

    def process_response(self, response) -> ModelExport:
        return ModelExport(**super().process_response(response)["createModelExport"])


class CreateModelExport(RequestChain):
    """
    Create a model export.

    Available on 6.14+ only.

    Args:
        model_id (int): the model id.
        wait (bool): wait for the export to complete. Defaults to True.
        request_interval (int | float): the interval between requests in seconds. Defaults to 5.
    """

    previous: Optional[ModelExport] = None

    def __init__(
        self,
        model_id: int,
        wait: bool = True,
        request_interval: Union[int, float] = 5,
    ):
        self.wait = wait
        self.model_id = model_id
        self.request_interval = request_interval
        super().__init__()

    def requests(self):
        yield _CreateModelExport(self.model_id)
        if self.wait:
            while self.previous and self.previous.status not in ["COMPLETE", "FAILED"]:
                yield GetModelExports([self.previous.id])
                self.previous = self.previous[0]
                yield Delay(seconds=self.request_interval)

        yield GetModelExports([self.previous.id], with_signed_url=self.wait is True)


class GetModelExports(GraphQLRequest):
    """
    Get model export(s).

    Available on 6.14+ only.

    Args:
        export_ids (list[int]): the export ids.
        with_signed_url (bool): include the signed url. Defaults to False
    """

    query = """
        query getModelExports($exportIds: [Int]) {
            modelExports(exportIds: $exportIds) {
                modelExports {
                    {fields}
                }
            }
        }
    """

    _base_fields = [
        "id",
        "name",
        "status",
        "modelId",
        "filePath",
        "createdAt",
        "createdBy",
    ]

    def __init__(self, export_ids: List[int], with_signed_url: bool = False):
        if with_signed_url:
            self._base_fields.append("signedUrl")

        query_with_fields = self.query.replace("{fields}", "\n".join(self._base_fields))
        super().__init__(query_with_fields, variables={"exportIds": export_ids})

    def process_response(self, response) -> List[ModelExport]:
        return [
            ModelExport(**export)
            for export in super().process_response(response)["modelExports"][
                "modelExports"
            ]
        ]



import json
import os
import time
from pathlib import Path

import pytest

from indico.client import IndicoClient
from indico.errors import IndicoRequestError
from indico.queries.jobs import JobStatus
from indico.queries.model_groups import (
    AddModelGroupComponent,
    GetModelGroup,
    GetTrainingModelWithProgress,
    ModelGroupPredict,
    UpdateModelGroupSettings,
)
from indico.queries.model_groups.metrics import (
    AnnotationModelGroupMetrics,
    GetModelGroupMetrics,
    ObjectDetectionMetrics,
)
from indico.queries.storage import URL_PREFIX, UploadDocument
from indico.types import Workflow
from indico.types.dataset import Dataset
from indico.types.model import Model, ModelOptions, TrainingProgress
from indico.types.model_group import ModelGroup

from ..data.datasets import (
    airlines_dataset,
    airlines_model_group,
    airlines_workflow,
    cats_dogs_image_dataset,
    cats_dogs_image_workflow,
    cats_dogs_modelgroup,
    org_annotate_dataset,
    org_annotate_model_group,
    org_annotate_workflow,
    too_small_dataset,
    too_small_workflow,
)


def test_get_missing_model_group(indico):
    client = IndicoClient()

    with pytest.raises(IndicoRequestError):
        client.call(GetModelGroup(id=500000))


def test_create_object_detection(
    cats_dogs_image_dataset: Dataset, cats_dogs_image_workflow: Workflow
):
    client = IndicoClient()
    name = f"TestCreateObjectDetectionMg-{int(time.time())}"

    model_training_options = {
        "max_iter": 2000,
        "lr": 0.1,
        "batch_size": 1,
        "filter_empty": False,
        "test_size": 0.2,
        "use_small_model": False,
    }
    after_component_id = cats_dogs_image_workflow.component_by_type("INPUT_IMAGE").id
    mg: ModelGroup = client.call(
        AddModelGroupComponent(
            name=name,
            dataset_id=cats_dogs_image_dataset.id,
            after_component_id=after_component_id,
            source_column_id=cats_dogs_image_dataset.datacolumn_by_name("urls").id,
            labelset_column_id=cats_dogs_image_dataset.labelset_by_name("label").id,
            workflow_id=cats_dogs_image_workflow.id,
            model_training_options=model_training_options,
        )
    )

    assert mg.model_group_by_name(name) is not None


def test_model_group_progress(
    indico, airlines_dataset: Dataset, airlines_workflow: Workflow
):
    client = IndicoClient()

    name = f"TestCreateModelGroup-{int(time.time())}"
    after_component_id = airlines_workflow.component_by_type("INPUT_OCR_EXTRACTION").id
    mg: ModelGroup = client.call(
        AddModelGroupComponent(
            name=name,
            dataset_id=airlines_dataset.id,
            after_component_id=after_component_id,
            source_column_id=airlines_dataset.datacolumn_by_name("Text").id,
            labelset_column_id=airlines_dataset.labelset_by_name("Target_1").id,
            workflow_id=airlines_workflow.id,
        )
    )
    time.sleep(1)
    model: Model = client.call(
        GetTrainingModelWithProgress(id=mg.model_group_by_name(name).id)
    )

    assert type(model) == Model
    assert model.status in ["CREATING", "TRAINING", "COMPLETE"]
    assert type(model.training_progress) == TrainingProgress
    assert model.training_progress.percent_complete < 101.0


def test_model_group_progress_bad_model_group_id(indico, airlines_dataset: Dataset):
    client = IndicoClient()

    with pytest.raises(IndicoRequestError):
        client.call((GetTrainingModelWithProgress(id=-1)))


def test_predict(indico, airlines_dataset, airlines_model_group):
    client = IndicoClient()

    job = client.call(
        ModelGroupPredict(
            model_id=airlines_model_group.selected_model.id,
            data=["I hate this airline"],
        )
    )

    assert type(job.id) == str

    job = client.call(JobStatus(id=job.id, wait=True))
    assert len(job.result) == 1


def get_storage_urls_from_fnames(client, image_fnames):
    client = IndicoClient()
    base_dir = Path(__file__).parent.parent / "data"
    image_filepaths = [os.path.join(base_dir, fname) for fname in image_fnames]
    response = client.call(UploadDocument(files=image_filepaths))
    storage_urls = [URL_PREFIX + json.loads(r["filemeta"])["path"] for r in response]
    return storage_urls


def test_object_detection_predict(indico, cats_dogs_modelgroup):
    client = IndicoClient()
    storage_urls = get_storage_urls_from_fnames(
        client, ["1.jpg", "2.jpg", "3.jpg", "4.jpg", "5.jpg"]
    )
    job = client.call(
        ModelGroupPredict(
            model_id=cats_dogs_modelgroup.selected_model.id,
            data=storage_urls,
            predict_options={"threshold": 0.25, "crops": True},
            load=False,
        )
    )
    assert type(job.id) == str

    result = client.call(JobStatus(job.id, wait=True))

    assert result.status != "FAILURE"
    assert len(result.result) == 5


def test_annotation_metrics(
    indico, org_annotate_dataset, org_annotate_workflow, org_annotate_model_group
):
    client = IndicoClient()
    result = client.call(
        AnnotationModelGroupMetrics(model_group_id=org_annotate_model_group.id)
    )
    check_annotation_metrics(result)


def test_object_detection_metrics(
    indico, cats_dogs_image_dataset, cats_dogs_image_workflow, cats_dogs_modelgroup
):
    client = IndicoClient()
    result = client.call(ObjectDetectionMetrics(cats_dogs_modelgroup.id))
    for metric_type in ["AP", "AP-Cat", "AP-Dog", "AP50", "AP75", "APl"]:
        assert isinstance(result["bbox"][metric_type], float)


def test_model_group_metrics_query(
    indico, org_annotate_dataset, org_annotate_model_group
):
    """This test is sometimes flaky"""
    client = IndicoClient()
    result = client.call(
        GetModelGroupMetrics(model_group_id=org_annotate_model_group.id)
    )
    check_annotation_metrics(result)


def check_annotation_metrics(result):
    assert result.class_metrics[0].name == "org"
    for attr in [
        "f1_score",
        "false_negatives",
        "false_positives",
        "precision",
        "recall",
        "true_positives",
    ]:
        assert isinstance(
            getattr(result.class_metrics[0].metrics[0], attr), (float, int)
        )
    assert isinstance(result.class_metrics[0].metrics[0].span_type, str)

    assert isinstance(result.model_level_metrics[0].span_type, str)
    assert isinstance(result.model_level_metrics[0].micro_f1, float)
    assert isinstance(result.model_level_metrics[0].macro_f1, float)
    assert isinstance(result.model_level_metrics[0].weighted_f1, float)
    assert result.retrain_for_metrics is False


def test_update_model_group_settings(indico, org_annotate_model_group):
    client = IndicoClient()
    model_training_options = {"use_autolabeled_data": True, "use_partial_data": True}
    result = client.call(
        UpdateModelGroupSettings(
            model_group_id=org_annotate_model_group.id,
            model_training_options=model_training_options,
        )
    )
    assert isinstance(result, ModelOptions)
    assert result.model_training_options == {
        "use_autolabeled_data": True,
        "use_partial_data": True,
    }



import datetime

from indico.types.base import BaseType


class ModelExport(BaseType):
    """
    The returned model export object.
    """

    id: int
    name: str
    status: str
    model_id: int
    file_path: str
    created_at: datetime.datetime
    created_by: int
    signed_url: str



import pytest
from indico.client import IndicoClient
from indico.filters import or_, UserMetricsFilter
from indico.queries import JobStatus, RetrieveStorageObject
from indico.types.user_metrics import UserSummary
from indico.queries.usermetrics import (
    GetUserSummary,
    GetUserSnapshots,
    GetUserChangelog,
    GenerateChangelogReport,
)
from datetime import datetime


def test_fetch_summary(indico):
    client = IndicoClient()
    user_summary: UserSummary = client.call(GetUserSummary())
    assert user_summary is not None
    assert len(user_summary.app_roles) > 0


def test_fetch_snapshots(indico):
    client = IndicoClient()
    snapshots = []
    for snapshot in client.paginate(GetUserSnapshots(date=datetime.now())):
        snapshots.extend(snapshot)
    assert snapshots is not None
    assert snapshots[0] is not None
    assert len(snapshots) > 0
    first = snapshots.pop()
    assert first is not None
    assert first.roles is not None


def test_fetch_snapshots_with_limit(indico):
    client = IndicoClient()
    snapshots = client.call(GetUserSnapshots(date=datetime.now(), limit=10))
    assert snapshots is not None
    assert snapshots[0] is not None
    assert len(snapshots) == 10
    first = snapshots.pop()
    assert first is not None
    assert first.roles is not None


def test_fetch_filtered_snapshots(indico):
    client = IndicoClient()
    snapshots = client.call(GetUserSnapshots(date=datetime.now(), limit=2))
    user_ids = [s.id for s in snapshots]
    filtered_snapshots = []
    filter_opts = or_(
        UserMetricsFilter(user_id=user_ids[0]), UserMetricsFilter(user_id=user_ids[1])
    )
    for snapshot in client.paginate(
        GetUserSnapshots(date=datetime.now(), filters=filter_opts)
    ):
        filtered_snapshots.extend(snapshot)
    assert len(filtered_snapshots) == 2


def test_changelog(indico):
    client = IndicoClient()
    changelogs = []
    for log in client.paginate(
        (
            GetUserChangelog(
                start_date=datetime.now(), end_date=datetime.now(), limit=100
            )
        )
    ):
        changelogs.extend(log)
    assert len(changelogs) > 0


def test_csv_changelog(indico):
    client = IndicoClient()
    changelogs = client.call(
        (GenerateChangelogReport(start_date=datetime.now(), end_date=datetime.now()))
    )
    assert changelogs is not None
    job = changelogs.job_id
    assert job is not None
    job = client.call(JobStatus(id=job, wait=True))
    assert job.status == "SUCCESS"
    assert job.ready is True
    result = client.call(RetrieveStorageObject(job.result))
    assert result is not None



import jsons
from typing import List

from indico import GraphQLRequest, RequestChain
from indico.errors import IndicoInputError
from indico.types.integration import (
    ExchangeIntegration,
    ExchangeIntegrationConfiguration,
    ExchangeIntegrationCredentials,
)


class AddExchangeIntegration(GraphQLRequest):
    """
    Mutation to add a Microsoft Exchange integration to a workflow

    Args:
        workflow_id(int): workflow to add integration to
        config(ExchangeIntegrationConfiguration): settings for which mailbox to point to and which emails to process
        credentials(ExchangeIntegrationCredentials): client id, client secret, and tenant id for authenticating with Exchange

    """

    query = """
    mutation addExchangeIntegration($workflow_id: Int!, $config: ExchangeIntegrationConfigurationInput!, $credentials: ExchangeIntegrationCredentialsInput!){
  addExchangeIntegrationToWorkflow(workflowId: $workflow_id, config: $config, credentials: $credentials){
    integration{
      id
      enabled
      workflowId
      createdAt
      config {
        filters
        userId
        folderName
        folderId
      }
    }
  }
}
    """

    def __init__(
        self,
        config: ExchangeIntegrationConfiguration,
        credentials: ExchangeIntegrationCredentials,
        workflow_id: int,
    ):
        super().__init__(
            self.query,
            variables={
                "config": config,
                "credentials": credentials,
                "workflow_id": workflow_id,
            },
        )

    def process_response(self, response) -> ExchangeIntegration:
        return ExchangeIntegration(
            **super().process_response(response)["addExchangeIntegrationToWorkflow"][
                "integration"
            ]
        )


class StartIntegration(GraphQLRequest):
    """
    Mutation to start an existing integration. Once an integration is started, documents will be submitted to the associated workflow.

    Args:
        integration_id(int): id of the integration to start

    """

    query = """
        mutation StartIntegration($integration_id: Int!){
          startWorkflowIntegration(integrationId: $integration_id){
            success
          }
        }
    """

    def __init__(
        self,
        integration_id: int,
    ):
        super().__init__(
            self.query,
            variables={"integration_id": integration_id},
        )


class DeleteIntegration(GraphQLRequest):
    """
    Mutation to delete an existing Integration.

    Args:
        integration_id(int): id of the integration to delete

    """

    query = """
        mutation DeleteWorkflowIntegration($integrationId: Int!) {
            deleteWorkflowIntegration(integrationId: $integrationId) {
                success
            }
        }
    """

    def __init__(self, integration_id: int):
        super().__init__(
            self.query,
            variables={"integrationId": integration_id},
        )


class PauseIntegration(GraphQLRequest):
    """
    Mutation to pause an existing Integration.

    Args:
        integration_id(int): id of the integration to pause
    """
    
    query = """
        mutation PauseIntegration($integrationId: Int!){
            pauseWorkflowIntegration(integrationId: $integrationId){
             success # bool
        }
    }
    """
    
    def __init__(self, integration_id: int):
        super().__init__(
            self.query,
            variables={"integrationId": integration_id},
        )



import pytest

from indico.client import AsyncIndicoClient, IndicoConfig
from indico.client.request import HTTPRequest, HTTPMethod, GraphQLRequest
from indico.errors import IndicoAuthenticationFailed


pytestmark = pytest.mark.asyncio(scope="module")


async def test_http_request(indico):
    async with AsyncIndicoClient() as client:
        response = await client.call(
            HTTPRequest(method=HTTPMethod.GET, path="/auth/users/details")
        )
        assert "email" in response


async def test_bad_token(indico):
    with pytest.raises(IndicoAuthenticationFailed):
        await AsyncIndicoClient(IndicoConfig(api_token="askldjflkasdjf")).create()


async def test_graphql_request(indico):
    async with AsyncIndicoClient() as client:
        response = await client.call(
            GraphQLRequest(
                query="""
        query {
            modelGroups{
                modelGroups{
                    id
                }
            }
        }
        """
            )
        )
    assert "modelGroups" in response


async def test_graphql_with_ids(aio_org_annotate_dataset):
    async with AsyncIndicoClient() as client:
        response = await client.call(
            GraphQLRequest(
                query="""
                query datasetQueries($id: Int!) {
                    dataset(id: $id){
                        id
                    }
                }
            """,
                variables={"id": aio_org_annotate_dataset.id},
            )
        )

    assert "dataset" in response



import io
import json
import tempfile
from typing import Dict, List, Union

from indico.client.request import Delay, GraphQLRequest, RequestChain
from indico.errors import IndicoError, IndicoInputError
from indico.queries.storage import UploadBatched, UploadDocument
from indico.types import SUBMISSION_RESULT_VERSIONS, Job, Submission, Workflow
from indico.types.utils import cc_to_snake, snake_to_cc


class ListWorkflows(GraphQLRequest):
    """
    List all workflows visible to authenticated user

    Options:
        dataset_ids (List[int]): List of dataset ids to filter by
        workflow_ids (List[int]): List of workflow ids to filter by

    Returns:
        List[Workflow]: All the found Workflow objects
    """

    query = """
        query ListWorkflows($datasetIds: [Int], $workflowIds: [Int], $limit: Int){
            workflows(datasetIds: $datasetIds, workflowIds: $workflowIds, limit: $limit){
                workflows {
                    id
                    name
                    status
                    reviewEnabled
                    autoReviewEnabled
                    createdAt
                    createdBy
                    submissionRunnable
                components {
                        id
                        componentType
                        reviewable
                        filteredClasses
                        ... on ContentLengthComponent
                        {
                                minimum
                                maximum
                                }
                        ... on ModelGroupComponent {
                            taskType
                            modelType
                            modelGroup {
                                        status
                                      id
                                      name
                                      taskType
                                      questionnaireId
                                      classNames
                                      selectedModel{
                                        id
                                      }
                                }

                        }

                    }
                  componentLinks{
                    id
                    headComponentId
                    tailComponentId
                    filters{
                      classes
                    }

                  }
                }
            }
        }
    """

    def __init__(
        self,
        *,
        dataset_ids: List[int] = None,
        workflow_ids: List[int] = None,
        limit=100,
    ):
        super().__init__(
            self.query,
            variables={
                "datasetIds": dataset_ids,
                "workflowIds": workflow_ids,
                "limit": limit,
            },
        )

    def process_response(self, response) -> List[Workflow]:
        return [
            Workflow(**w)
            for w in super().process_response(response)["workflows"]["workflows"]
        ]


class GetWorkflow(ListWorkflows):
    """
    Query for Workflow by id

    Args:
        workflow_id (int): Workflow id to query for

    Returns:
        Workflow: Found Workflow object
    """

    def __init__(self, workflow_id: int):
        super().__init__(workflow_ids=[workflow_id])

    def process_response(self, response) -> Workflow:
        return super().process_response(response)[0]


class _ToggleReview(GraphQLRequest):
    toggle = "enableReview"
    query_name = "toggleWorkflowReview"
    query = """
        mutation ToggleReview($workflowId: Int!, $reviewState: Boolean!){
            <QUERY NAME>(workflowId: $workflowId, <TOGGLE>: $reviewState){
                id
                name
                reviewEnabled
                autoReviewEnabled
            }
        }
    """

    def __init__(self, workflow_id: int, enable_review: bool):
        query = self.query.replace("<QUERY NAME>", self.query_name)
        query = query.replace("<TOGGLE>", self.toggle)
        super().__init__(
            query,
            variables={"workflowId": workflow_id, "reviewState": enable_review},
        )

    def process_response(self, response) -> Workflow:
        return Workflow(**super().process_response(response)[self.query_name])


class _ToggleAutoReview(_ToggleReview):
    toggle = "enableAutoReview"
    query_name = "toggleWorkflowAutoReview"


class UpdateWorkflowSettings(RequestChain):
    """
    Mutation to toggle review and auto-review on a workflow

    Args:
        workflow (int|Workflow): Workflow or workflow id to update

    Options:
        enable_review (bool): Enable review
        enable_auto_review (bool): Enable auto_review

    Returns:
        Workflow: Updated Workflow object
    """

    def __init__(
        self,
        workflow: Union[Workflow, int],
        enable_review: bool = None,
        enable_auto_review: bool = None,
    ):
        self.workflow_id = workflow.id if isinstance(workflow, Workflow) else workflow
        if enable_review is None and enable_auto_review is None:
            raise IndicoInputError("Must provide at least one review option")

        self.enable_review = enable_review
        self.enable_auto_review = enable_auto_review

    def requests(self):
        if self.enable_review is not None:
            yield _ToggleReview(self.workflow_id, self.enable_review)
        if self.enable_auto_review is not None:
            yield _ToggleAutoReview(self.workflow_id, self.enable_auto_review)


class _WorkflowSubmission(GraphQLRequest):
    query = """
        mutation workflowSubmissionMutation({signature}) {{
            {mutation_name}({args}) {{
                jobIds
                submissionIds
            }}
        }}
    """

    detailed_query = """
        mutation workflowSubmissionMutation({signature}) {{
            {mutation_name}({args}) {{
                submissionIds
                submissions {{
                    id
                    datasetId
                    workflowId
                    status
                    <SUBQUERY>
                    inputFile
                    inputFilename
                    resultFile
                    retrieved
                    errors
                }}
            }}
        }}
    """
    files_subquery = """
        inputFiles {{
            filepath
            filename
        }}
    """.strip()

    mutation_name = "workflowSubmission"
    mutation_args = {
        "workflowId": "Int!",
        "files": "[FileInput]!",
        "bundle": "Boolean",
        "resultVersion": "SubmissionResultVersion",
    }

    def __init__(
        self,
        detailed_response: bool,
        **kwargs,
    ):
        self.workflow_id = kwargs["workflow_id"]

        # construct mutation signature and args based on provided kwargs to ensure
        # backwards-compatible graphql calls
        #
        # inputFiles, bundle, and resultVersion only avaliable on IPA 4.9.0+
        subq = (
            self.files_subquery
            if kwargs.get("bundle") or kwargs.get("result_version")
            else ""
        )
        q = (
            self.detailed_query.replace("<SUBQUERY>", subq)
            if detailed_response
            else self.query
        )

        args = [
            _arg for _arg in self.mutation_args.keys() if kwargs.get(cc_to_snake(_arg))
        ]
        signature = ",".join(f"${_arg}: {self.mutation_args[_arg]}" for _arg in args)
        args = ",".join(f"{_arg}: ${_arg}" for _arg in args)

        super().__init__(
            query=q.format(
                mutation_name=self.mutation_name, signature=signature, args=args
            ),
            variables={snake_to_cc(var): val for var, val in kwargs.items()},
        )

    def process_response(self, response):
        response = super().process_response(response)[self.mutation_name]
        if "submissions" in response:
            return [Submission(**s) for s in response["submissions"]]
        if not response["submissionIds"]:
            raise IndicoError(f"Failed to submit to workflow {self.workflow_id}")
        else:
            return response["submissionIds"]
        return [Job(id=job_id) for job_id in response["jobIds"]]


class _WorkflowUrlSubmission(_WorkflowSubmission):
    mutation_name = "workflowUrlSubmission"
    mutation_args = {**_WorkflowSubmission.mutation_args, "urls": "[String]!"}
    del mutation_args["files"]


class WorkflowSubmission(RequestChain):
    f"""
    Submit files to a workflow for processing.
    One of `files`, `urls`, `stream`, or `text` is required.

    Args:
        workflow_id (int): Id of workflow to submit files to
        files (List[str], optional): List of local file paths to submit
        urls (List[str], optional): List of urls to submit
        submission (bool, optional): DEPRECATED - AsyncJobs are no longer supported.
        bundle (bool, optional): Batch all files under a single submission id
        result_version (str, optional):
            The format of the submission result file. One of:
                {SUBMISSION_RESULT_VERSIONS}
            If bundle is enabled, this must be version TWO or later.
        streams (Dict[str, io.BufferedIOBase]): List of filename keys mapped to streams
            for upload. Similar to files but mutually exclusive with files.
            Can take for example: io.BufferedReader, io.BinaryIO, or io.BytesIO.
        text (str, optional): text to submit. Note: submission may still go through OCR.
        batch_size (int, optional): If submitting files, specifies the amount of files to submit in a single batch. Defaults to 10. A call with a batch exceeding 500mb total will fail with an error.

    Returns:
        List[int]: If `submission`, these will be submission ids.
            Otherwise, they will be AsyncJob ids.

    """

    detailed_response = False

    def __init__(
        self,
        workflow_id: int,
        files: List[str] = None,
        urls: List[str] = None,
        submission: bool = True,
        bundle: bool = False,
        result_version: str = None,
        streams: Dict[str, io.BufferedIOBase] = None,
        text: str = "",
        batch_size: int = 10,
    ):
        self.workflow_id = workflow_id
        self.files = files
        self.urls = urls
        self.submission = submission
        self.bundle = bundle
        self.result_version = result_version
        self.has_streams = False
        if streams is not None:
            self.streams = streams.copy()
            self.has_streams = True
        else:
            self.streams = None
        self.text = text
        self.batch_size = batch_size
        if not submission:
            raise IndicoInputError("This option is deprecated and no longer supported.")
        if not self.files and not self.urls and not self.has_streams and not self.text:
            raise IndicoInputError(
                "One of 'files', 'streams', 'urls', or 'text' must be specified"
            )
        elif (self.files or self.urls or self.text) and self.has_streams:
            raise IndicoInputError(
                "Only one of 'files' or 'streams', 'urls', or 'text' may be specified."
            )
        elif (self.files or self.text) and self.urls:
            raise IndicoInputError(
                "Only one of 'files' or 'streams', 'urls', or 'text' may be specified"
            )
        elif self.files and self.text:
            raise IndicoInputError(
                "Only one of 'files' or 'streams', 'urls', or 'text' may be specified"
            )

    def requests(self):
        if self.files:
            yield UploadBatched(files=self.files, batch_size=self.batch_size)
            yield _WorkflowSubmission(
                self.detailed_response,
                workflow_id=self.workflow_id,
                files=self.previous,
                bundle=self.bundle,
                result_version=self.result_version,
            )
        elif self.urls:
            yield _WorkflowUrlSubmission(
                self.detailed_response,
                workflow_id=self.workflow_id,
                urls=self.urls,
                bundle=self.bundle,
                result_version=self.result_version,
            )
        elif self.has_streams:
            yield UploadDocument(streams=self.streams)
            yield _WorkflowSubmission(
                self.detailed_response,
                workflow_id=self.workflow_id,
                files=self.previous,
                bundle=self.bundle,
                result_version=self.result_version,
            )
        elif self.text:
            temp = tempfile.NamedTemporaryFile(mode="w+", suffix=".txt")
            try:
                temp.write(self.text)
                temp.seek(0)
                yield UploadDocument(files=[temp.name])
                yield _WorkflowSubmission(
                    self.detailed_response,
                    workflow_id=self.workflow_id,
                    files=self.previous,
                    bundle=self.bundle,
                    result_version=self.result_version,
                )
            finally:
                temp.close()


class WorkflowSubmissionDetailed(WorkflowSubmission):
    f"""
    Submit files to a workflow for processing.
    One of `files` or `urls` is required.
    Submission recording is mandatory.

    Args:
        workflow_id (int): Id of workflow to submit files to
        files (List[str], optional): List of local file paths to submit
        urls (List[str], optional): List of urls to submit
        bundle (bool, optional): Batch all files under a single submission id
        result_version (str, optional):
            The format of the submission result file. One of:
                {SUBMISSION_RESULT_VERSIONS}
            If bundle is enabled, this must be version TWO or later.

    Returns:
        List[Submission]: Submission objects created

    """

    detailed_response = True

    def __init__(
        self,
        workflow_id: int,
        files: List[str] = None,
        urls: List[str] = None,
        bundle: bool = False,
        result_version: str = None,
    ):
        super().__init__(
            workflow_id,
            files=files,
            urls=urls,
            submission=True,
            bundle=bundle,
            result_version=result_version,
        )


class _AddDataToWorkflow(GraphQLRequest):
    query = """
        mutation addDataToWorkflow($workflowId: Int!) {
            addDataToWorkflow(workflowId: $workflowId){
                workflow{
                    id
                    name
                    status
                }
            }
        }
    """

    def __init__(self, workflow_id: int):
        super().__init__(
            self.query,
            variables={"workflowId": workflow_id},
        )

    def process_response(self, response) -> Workflow:
        return Workflow(
            **super().process_response(response)["addDataToWorkflow"]["workflow"]
        )


class AddDataToWorkflow(RequestChain):
    """
    Mutation to update data in a workflow, presumably
    after new data is added to the dataset.

    Args:
        workflow_id (int): Workflow id to update

    Options:
        wait (bool, default=False): Block while polling for status of update

    Returns:
        Workflow: Updated Workflow object
    """

    def __init__(self, workflow_id: int, wait: bool = False):
        self.workflow_id = workflow_id
        self.wait = wait

    def requests(self):
        yield _AddDataToWorkflow(self.workflow_id)

        if self.wait:
            while self.previous.status != "COMPLETE":
                yield GetWorkflow(workflow_id=self.workflow_id)
                yield Delay()


class CreateWorkflow(GraphQLRequest):
    """
    Mutation to create workflow given an existing dataset.

    Args:
        dataset_id(int): dataset to create workflow for
        name(str): name for the workflow

    """

    query = """  mutation createWorkflow($datasetId: Int!, $name: String!) {
    createWorkflow(datasetId: $datasetId, name: $name) {
           workflow {
                    id
                    name
                    status
                    reviewEnabled
                    autoReviewEnabled
                    submissionRunnable
                components {
                        id
                        componentType
                        reviewable
                        filteredClasses

                        ... on ModelGroupComponent {
                            taskType
                            modelType
                        }

                    }
                  componentLinks{
                    id
                    headComponentId
                    tailComponentId
                    filters {
                      classes
                    }

                  }
                }
    }
    }
    """

    def __init__(self, dataset_id: int, name: str):
        super().__init__(
            self.query,
            variables={"datasetId": dataset_id, "name": name},
        )

    def process_response(self, response) -> Workflow:
        return Workflow(
            **super().process_response(response)["createWorkflow"]["workflow"]
        )


class DeleteWorkflow(GraphQLRequest):
    """
    Mutation to delete workflow given workflow id. Note that this operation includes deleting
    all components and models associated with this workflow.

    Args:
        workflow_id(int): id of workflow to delete
    """

    query = """
        mutation deleteWorkflow($workflowId: Int!){
            deleteWorkflow(workflowId: $workflowId){
                success
            }
        }
    """

    def __init__(self, workflow_id: int):
        super().__init__(self.query, variables={"workflowId": workflow_id})

    def process_response(self, response) -> bool:
        return super().process_response(response)["deleteWorkflow"]["success"]


import os
import pytest
import time
from pathlib import Path
from indico.client import IndicoClient
from indico.queries import (
    CreateDataset,
    CreateWorkflow,
    AddModelGroupComponent,
    GetModelGroup,
    AddExchangeIntegration,
    StartIntegration,
    GetWorkflow,
)
from indico.queries.workflow_components import _AddWorkflowComponent
from indico.types import ModelGroup, Dataset, Workflow, Integration

PUBLIC_URL = "https://github.com/IndicoDataSolutions/indico-client-python/raw/master/tests/integration/data/"


@pytest.fixture(scope="module")
def airlines_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[0]) + "/AirlineComplaints.csv"

    response = client.call(
        CreateDataset(
            name=f"AirlineComplaints-test-{int(time.time())}", files=[dataset_filepath]
        )
    )
    assert response.status == "COMPLETE"
    return response


@pytest.fixture(scope="module")
def airlines_workflow(indico, airlines_dataset: Dataset):
    client = IndicoClient()
    workflowreq = CreateWorkflow(
        dataset_id=airlines_dataset.id,
        name=f"AirlineComplaints-test-{int(time.time())}",
    )
    wf = client.call(workflowreq)
    # add default output node
    client.call(
        _AddWorkflowComponent(
            after_component_id=wf.component_by_type("OUTPUT_JSON_FORMATTER").id,
            component='{"component_type":"default_output","config":{}}',
            workflow_id=wf.id,
            after_component_link=None,
        )
    )
    response = client.call(GetWorkflow(workflow_id=wf.id))

    return response


@pytest.fixture(scope="module")
def too_small_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[0]) + "/TooSmall.csv"

    response = client.call(
        CreateDataset(
            name=f"TooSmall-test-{int(time.time())}", files=[dataset_filepath]
        )
    )
    assert response.status == "COMPLETE"
    return response


@pytest.fixture(scope="module")
def too_small_workflow(indico, too_small_dataset: Dataset) -> Workflow:
    client = IndicoClient()
    workflowreq = CreateWorkflow(
        dataset_id=too_small_dataset.id, name=f"TooSmall-test-{int(time.time())}"
    )
    wf = client.call(workflowreq)
    # add default output node
    client.call(
        _AddWorkflowComponent(
            after_component_id=wf.component_by_type("OUTPUT_JSON_FORMATTER").id,
            component='{"component_type":"default_output","config":{}}',
            workflow_id=wf.id,
            after_component_link=None,
        )
    )
    response = client.call(GetWorkflow(workflow_id=wf.id))

    return response


@pytest.fixture(scope="module")
def airlines_model_group(
    indico, airlines_dataset: Dataset, airlines_workflow: Workflow
) -> ModelGroup:
    client = IndicoClient()
    name = f"TestCreateModelGroup-{int(time.time())}"
    after_component_id = airlines_workflow.component_by_type("INPUT_OCR_EXTRACTION").id
    modelgroupreq = AddModelGroupComponent(
        name=name,
        dataset_id=airlines_dataset.id,
        after_component_id=after_component_id,
        source_column_id=airlines_dataset.datacolumn_by_name("Text").id,
        labelset_column_id=airlines_dataset.labelset_by_name("Target_1").id,
        workflow_id=airlines_workflow.id,
    )

    workflow_update: Workflow = client.call(modelgroupreq)
    component = workflow_update.model_group_by_name(name)

    mg = client.call(GetModelGroup(id=component.model_group.id, wait=True))

    return mg


@pytest.fixture(scope="module")
def cats_dogs_image_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[0]) + "/dog_vs_cats_small.csv"

    response = client.call(
        CreateDataset(
            name=f"DogsAndCats-test-{int(time.time())}",
            files=dataset_filepath,
            from_local_images=True,
        )
    )
    assert response.status == "COMPLETE"
    return response


@pytest.fixture(scope="module")
def cats_dogs_image_workflow(indico, cats_dogs_image_dataset: Dataset):
    client = IndicoClient()
    workflowreq = CreateWorkflow(
        dataset_id=cats_dogs_image_dataset.id,
        name=f"DogsAndCats-test-{int(time.time())}",
    )
    wf = client.call(workflowreq)
    client.call(
        _AddWorkflowComponent(
            after_component_id=wf.component_by_type("OUTPUT_JSON_FORMATTER").id,
            component='{"component_type":"default_output","config":{}}',
            workflow_id=wf.id,
            after_component_link=None,
        )
    )

    return wf


@pytest.fixture(scope="module")
def cats_dogs_modelgroup(
    indico, cats_dogs_image_dataset: Dataset, cats_dogs_image_workflow
) -> ModelGroup:
    client = IndicoClient()
    name = f"TestCreateObjectDetectionMg-{int(time.time())}"

    model_training_options = {
        "max_iter": 20,
        "lr": 0.1,
        "batch_size": 1,
        "filter_empty": False,
        "test_size": 0.2,
        "use_small_model": True,
    }
    after_component_id = cats_dogs_image_workflow.component_by_type("INPUT_IMAGE").id
    modelgroupreq = AddModelGroupComponent(
        name=name,
        dataset_id=cats_dogs_image_dataset.id,
        after_component_id=after_component_id,
        source_column_id=cats_dogs_image_dataset.datacolumn_by_name("urls").id,
        labelset_column_id=cats_dogs_image_dataset.labelset_by_name("label").id,
        workflow_id=cats_dogs_image_workflow.id,
        model_training_options=model_training_options,
    )

    workflow_update: Workflow = client.call(modelgroupreq)
    component = workflow_update.model_group_by_name(name)
    mg = client.call(GetModelGroup(id=component.model_group.id, wait=True))

    return mg


@pytest.fixture(scope="module")
def org_annotate_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[0]) + "/org-annotate-labeled.csv"

    response = client.call(
        CreateDataset(
            name=f"OrgAnnotate-test-{int(time.time())}", files=[dataset_filepath]
        )
    )
    assert response.status == "COMPLETE"
    return response


@pytest.fixture(scope="module")
def org_annotate_workflow(indico, org_annotate_dataset: Dataset):
    client = IndicoClient()
    workflowreq = CreateWorkflow(
        dataset_id=org_annotate_dataset.id, name=f"OrgAnnotate-test-{int(time.time())}"
    )
    wf = client.call(workflowreq)
    # add default output node
    client.call(
        _AddWorkflowComponent(
            after_component_id=wf.component_by_type("OUTPUT_JSON_FORMATTER").id,
            component='{"component_type":"default_output","config":{}}',
            workflow_id=wf.id,
            after_component_link=None,
        )
    )
    response = client.call(GetWorkflow(workflow_id=wf.id))
    return response


@pytest.fixture(scope="module")
def org_annotate_model_group(
    indico, org_annotate_dataset: Dataset, org_annotate_workflow: Workflow
) -> ModelGroup:
    client = IndicoClient()
    name = f"TestFinetuneModelGroup-{int(time.time())}"
    after_component_id = org_annotate_workflow.component_by_type(
        "INPUT_OCR_EXTRACTION"
    ).id
    modelgroupreq = AddModelGroupComponent(
        name=name,
        dataset_id=org_annotate_dataset.id,
        after_component_id=after_component_id,
        source_column_id=org_annotate_dataset.datacolumn_by_name(
            "News Headlines w/Company Names"
        ).id,
        labelset_column_id=org_annotate_dataset.labelset_by_name("question_825").id,
        workflow_id=org_annotate_workflow.id,
    )

    workflow_update: Workflow = client.call(modelgroupreq)
    component = workflow_update.model_group_by_name(name)

    mg = client.call(GetModelGroup(id=component.model_group.id, wait=True))

    return mg


@pytest.fixture(scope="module")
def org_annotate_exchange_integration(org_annotate_workflow: Workflow) -> Integration:
    client = IndicoClient()
    creds = {
        "clientId": os.getenv("EXCH_CLIENT_ID"),
        "clientSecret": os.getenv("EXCH_CLIENT_SECRET"),
        "tenantId": os.getenv("EXCH_TENANT_ID"),
    }

    config = {"userId": os.getenv("EXCH_USER_ID"), "folderId": "mailFolders('inbox')"}

    integ: Integration = client.call(
        AddExchangeIntegration(
            workflow_id=org_annotate_workflow.id, config=config, credentials=creds
        )
    )

    return integ


@pytest.fixture(scope="module")
def exchange_integration_to_delete(org_annotate_workflow: Workflow) -> Integration:
    client = IndicoClient()
    creds = {
        "clientId": os.getenv("EXCH_CLIENT_ID"),
        "clientSecret": os.getenv("EXCH_CLIENT_SECRET"),
        "tenantId": os.getenv("EXCH_TENANT_ID")
    }

    config = {
        "userId": os.getenv("EXCH_USER_ID"),
        "folderId": "mailFolders('inbox')"
    }

    integ: Integration = client.call(
        AddExchangeIntegration(
            workflow_id=org_annotate_workflow.id,
            config=config,
            credentials=creds
        )
    )

    return integ


@pytest.fixture(scope="module")
def started_exchange_integration(org_annotate_workflow: Workflow) -> Integration:
    client = IndicoClient()
    creds = {
        "clientId": os.getenv("EXCH_CLIENT_ID"),
        "clientSecret": os.getenv("EXCH_CLIENT_SECRET"),
        "tenantId": os.getenv("EXCH_TENANT_ID")
    }

    config = {
        "userId": os.getenv("EXCH_USER_ID"),
        "folderId": "mailFolders('inbox')"
    }

    integ: Integration = client.call(
        AddExchangeIntegration(
            workflow_id=org_annotate_workflow.id,
            config=config,
            credentials=creds
        )
    )
    client.call(
        StartIntegration(
            integration_id=integ.id
        )
    )
    yield integ



from typing import Dict, List, Union

from indico.client.request import GraphQLRequest, RequestChain, PagedRequest
from indico.filters import ModelGroupExampleFilter
from indico.types import Example, model_group


class ListModelGroupExamples(PagedRequest):
    """
    List all examples associated with a given model group ID.
    Supports pagination (limit becomes page_size)

    Options:
        model_group_ids (List[int]): Model group ids to filter by
        filters (ModelGroupExampleFilter or Dict): Example attributes to filter by
        limit (int, default=1000): Maximum number of Examples to return
        orderBy (str, default="ID"): Example attribute to filter by
        desc: (bool, default=True): List in descending order

    Returns:
        List[Example]: All the found Example objects
        If paginated, yields results one at a time
    """

    query = """
        query GetExamples($modelGroupId:Int, $orderBy: ExampleOrder, $desc: Boolean, $limit: Int, $filters: ModelGroupExampleFilter, $after: Int, $before: Int) {
            modelGroups(modelGroupIds: [$modelGroupId]) {
                modelGroups {
                    pagedExamples(orderBy:$orderBy, desc:$desc, limit: $limit, filters: $filters, after: $after, before: $before) {
                        examples {
                            id
                            status
                            datafileId
                        }
                        pageInfo {
                            startCursor
                            endCursor
                            hasNextPage
                        }
                    }  
                }
            }                
        }
    """

    def __init__(
        self,
        *,
        model_group_id: int = None,
        filters: Union[Dict, ModelGroupExampleFilter] = None,
        limit: int = 1000,
        order_by: str = "ID",
        desc: bool = True,
        after: int = None,
        before: int = None,
    ):
        variables = {
            "modelGroupId": model_group_id,
            "filters": filters,
            "limit": limit,
            "orderBy": order_by,
            "desc": desc,
            "after": after,
            "before": before,
        }
        super().__init__(
            self.query,
            variables=variables,
        )

    def process_response(self, response) -> List[Example]:
        response = response["data"]["modelGroups"]["modelGroups"][0]
        _pg = next(iter(response.values()))["pageInfo"]
        self.has_next_page = _pg["hasNextPage"]
        self.variables["after"] = _pg["endCursor"] if self.has_next_page else None
        return [Example(**s) for s in response["pagedExamples"]["examples"]]



import typing as t

from indico import GraphQLRequest
from indico.errors import IndicoInputError
from indico.types.workflow import ComponentFamily
from indico.types.custom_blueprint import TaskBlueprint

SUPPORTED_CUSTOM_COMPONENT_FAMILIES = [
    ComponentFamily.OUTPUT,
    ComponentFamily.FILTER,
    ComponentFamily.MODEL,
]


class RegisterCustomBlueprint(GraphQLRequest):
    """
    Mutation to register a custom blueprint, making it available in the gallery to add to workflows

    Args:
        component_family (ComponentFamily): family this blueprint belongs to; supported families are Output, Filter, and Model
        name (str): Name of the blueprint
        description (str): Description of this blueprint
        config (dict): Blueprint configuration options
        tags: (list[str]): List of tags associated with this blueprint

    Options:
        footer (str): Footnote for this blueprint's description
        icon (str): Image to use when displaying this blueprint in the gallery. Can be the name of an Indico-provided image
        or the storage location of a custom image. If left unspecified, the platform will use a default icon.
        all_access (bool): This blueprint can be added to all workflows
        dataset_ids (list[int]): This blueprint can only be added to workflows associated with these dataset ids

    Returns:
        TaskBlueprint: The newly created TaskBlueprint object
    """

    query = """
mutation createCustomBP(
  $name: String!,
  $description: String!,
  $footer: String!,
  $tags: [BlueprintTag]!,
  $config: {config_type}!,
  $icon: String,
  $allAccess: Boolean,
  $datasetIds: [Int]
) {{
  {mutation_name}(
    name: $name,
    description: $description,
    footer: $footer,
    tags: $tags,
    config: $config,
    icon: $icon,
    allAccess: $allAccess,
    datasetIds: $datasetIds
  ) {{
    id
    name
    componentFamily
    icon
    description
    footer
    tags
    enabled
    config
  }}
}}
"""

    def __init__(
        self,
        component_family: ComponentFamily,
        name: str,
        description: str,
        config: t.Dict,
        tags: t.List[str],
        footer: str = "",
        icon: str = None,
        all_access: bool = None,
        dataset_ids: t.List[int] = None,
    ):
        if (
            not component_family
            or not name
            or not description
            or not config
            or not tags
        ):
            none_vals = []
            if not component_family:
                none_vals.append("component_family")
            if not name:
                none_vals.append("name")
            if not description:
                none_vals.append("description")
            if not config:
                none_vals.append("config")
            if not tags:
                none_vals.append("tags")
            raise IndicoInputError(
                f"The following arguments cannot be None: {', '.join(none_vals)}"
            )
        if not all([isinstance(t, str) for t in tags]):
            raise IndicoInputError("'tags' must be a list of strings")
        if not all(config.values()):
            raise IndicoInputError("values in the 'config' dict must not be None")
        if not isinstance(component_family, ComponentFamily):
            raise IndicoInputError(
                "'component_family' must be of type indico.types.workflows.ComponentFamily"
            )
        if component_family not in SUPPORTED_CUSTOM_COMPONENT_FAMILIES:
            raise IndicoInputError(
                f"component_family must be one of {', '.join([cf.name for cf in SUPPORTED_CUSTOM_COMPONENT_FAMILIES])}"
            )

        comp_fam_str = component_family.name.title()
        self.mutation_name = f"createCustom{comp_fam_str}TaskBlueprint"
        super().__init__(
            self.query.format(
                config_type=f"{comp_fam_str}Config",
                mutation_name=self.mutation_name,
            ),
            variables={
                "name": name,
                "description": description,
                "footer": footer,
                "config": config,
                "tags": tags,
                "icon": icon,
                "allAccess": all_access,
                "datasetIds": dataset_ids,
            },
        )

    def process_response(self, response) -> TaskBlueprint:
        return TaskBlueprint(**super().process_response(response)[self.mutation_name])



from enum import Enum
from typing import List, Union
from datetime import datetime
from indico.types import BaseType, ModelGroup, ModelTaskType


class WorkflowComponent(BaseType):
    """
    A component, such as a Model Group or Content Length filter, that is present on a workflow.
    This is essentially a step in the workflow process, such as OCR
    or predicting with a model group.

    """

    id: int
    component_type: str
    task_type: str
    model_type: str
    model_group: ModelGroup
    task_type: str
    model_type: str
    minimum: int
    maximum: int


class WorkflowComponentLinks(BaseType):
    """
    Represents a link between two components.
    """

    id: int
    head_component_id: int
    tail_component_id: int


class Workflow(BaseType):
    """
    Represents a Workflow in the Indico Data Platform.
    """

    id: int
    name: str
    status: str
    review_enabled: bool
    auto_review_enabled: bool
    components: List[WorkflowComponent]
    component_links: List[WorkflowComponentLinks]
    created_by: int
    created_at: datetime
    submission_runnable: bool

    def component_by_type(self, component_type: str) -> WorkflowComponent:
        """
        Returns first component available of type specified.
        """
        return next(
            component
            for component in self.components
            if component.component_type == component_type
        )

    def model_group_by_name(self, name: str) -> WorkflowComponent:
        """
        Returns first model group component of name specified.
        """
        return next(
            component
            for component in self.components
            if hasattr(component, "model_group") and component.model_group.name == name
        )


class LinkedLabelStrategy(Enum):
    BY_ROW = 0,
    BY_KEY = 1


class LinkedLabelGroup:
    def __init__(
        self,
        name: str,
        strategy: LinkedLabelStrategy,
        class_ids: List[int],
        strategy_settings: dict,
    ):
        self.name = name
        self.strategy = strategy
        self.class_ids = class_ids
        self.strategy_settings = strategy_settings


class ComponentFamily(Enum):
    MODEL = 0,
    FILTER = 1,
    TRANSFORMER = 2,
    REVIEW = 3,
    OUTPUT = 4



import json
from typing import Any, Dict, List, Optional, Union

import deprecation

from indico.client.request import Delay, GraphQLRequest, RequestChain
from indico.errors import IndicoNotFound
from indico.queries.workflow_components import AddModelGroupComponent
from indico.types.jobs import Job
from indico.types.model import Model, ModelOptions
from indico.types.model_group import ModelGroup
from indico.types.utils import cc_to_snake


class GetModelGroup(RequestChain):
    """
    Get an object describing a model group

    Args:
        id (int): model group id to query
        wait (bool, optional): Wait until the Model Group status is FAILED, COMPLETE, or NOT_ENOUGH_DATA. Defaults to False.
        request_interval (int or float, optional): The maximum time in between retry calls when waiting. Defaults to 5 seconds.

    Returns:
        ModelGroup object
    """

    def __init__(
        self, id: int, wait: bool = False, request_interval: Union[int, float] = 5
    ):
        self.id = id
        self.wait = wait
        self.request_interval = request_interval

    def requests(self):
        if self.wait:
            req = GetModelGroupSelectedModelStatus(id=self.id)
            yield req
            while self.previous not in ["FAILED", "COMPLETE", "NOT_ENOUGH_DATA"]:
                yield Delay(seconds=self.request_interval)
                yield req
        yield _GetModelGroup(id=self.id)


class _GetModelGroup(GraphQLRequest):
    """
    Get an object describing a model group

    Args:
        id (int): model group id to query

    Returns:
        ModelGroup object
    """

    query = """
        query GetModelGroup($id: Int){
            modelGroups(modelGroupIds: [$id]) {
                    modelGroups {
                        id
                        name
                        status
                        taskType
                        selectedModel {
                            id
                            status
                        }
                    }
                }
            }
        """

    def __init__(self, id: int):
        super().__init__(query=self.query, variables={"id": id})

    def process_response(self, response):
        try:
            mg = ModelGroup(
                **super().process_response(response)["modelGroups"]["modelGroups"][0]
            )
        except IndexError:
            raise IndicoNotFound(
                "ModelGroup not found. Please check the ID you are using."
            )
        return mg


class GetTrainingModelWithProgress(GraphQLRequest):
    """
    Get progress (percent complete) of a training model group

    Args:
        id (int): model group id to query

    Returns:
        Model object with percent_complete field
    """

    query = """
        query ModelGroupProgress($id: Int){
            modelGroups(modelGroupIds: [$id]){
                modelGroups{
                    models {
                        id
                        status
                        trainingProgress {
                            percentComplete
                        }
                    }
                }
            }
        }
    """

    def __init__(self, id: int):
        super().__init__(query=self.query, variables={"id": id})

    def process_response(self, response):
        response = super().process_response(response)
        model_groups = response["modelGroups"]["modelGroups"]
        if len(model_groups) != 1:
            raise IndicoNotFound("Model Group")
        models = model_groups[0]["models"]

        last = max(models, key=lambda m: m["id"])
        if not last:
            raise IndicoNotFound("Training Model")

        return Model(**last)


class GetModelGroupSelectedModelStatus(GraphQLRequest):
    """
    Get the status string of the selected model for the given model group id

    Args:
        id (int): model group id to query

    Returns:
        status (str): CREATED, TRAINING, COMPLETE or FAILED
    """

    query = """
        query GetModelGroup($id: Int) {
                modelGroups(modelGroupIds: [$id]) {
                    modelGroups {
                        id
                        selectedModel {
                            id
                            status
                        }
                    }
                }
            }
        """

    def __init__(self, id: int):
        super().__init__(query=self.query, variables={"id": id})

    def process_response(self, response):
        mg = ModelGroup(
            **super().process_response(response)["modelGroups"]["modelGroups"][0]
        )
        return mg.selected_model.status


class _ModelGroupPredict(GraphQLRequest):
    query = """
        mutation ModelGroupPredict(<QUERY_ARGS>) {
            modelPredict(<MODEL_PREDICT_ARGS>) {
                jobId
            }
        }
        """

    query_args = {"modelId": "Int!", "data": "[String]", "predictOptions": "JSONString"}

    def _args_strings(self, **kwargs):
        args = [k for k in self.query_args.keys() if kwargs.get(cc_to_snake(k))]

        query_args_string = ",".join(f"${k}: {self.query_args[k]}" for k in args)
        model_predict_args = ",".join(f"{k}: ${k}" for k in args)

        query = self.query.replace("<QUERY_ARGS>", query_args_string)
        query = query.replace("<MODEL_PREDICT_ARGS>", model_predict_args)

        return query

    def __init__(self, model_id: int, data: List[str], predict_options: Dict = None):
        if predict_options:
            predict_options = json.dumps(predict_options)

        query = self._args_strings(
            model_id=model_id, data=data, predict_options=predict_options
        )

        super().__init__(
            query=query,
            variables={
                "modelId": model_id,
                "data": data,
                "predictOptions": predict_options,
            },
        )

    def process_response(self, response):
        return Job(**super().process_response(response)["modelPredict"])


class ModelGroupPredict(RequestChain):
    """
    Generate predictions from a model group on new data

    Args:
        model_id (int): selected model id use for predictions
        data (List[str]): list of samples to predict
        predict_options (JSONString): arguments for predictions

    Returns:
        Job associated with this model group predict task
    """

    def __init__(
        self,
        model_id: int,
        data: List[str],
        load: bool = True,
        predict_options: Dict = None,
    ):
        self.model_id = model_id
        self.data = data
        self.predict_options = predict_options

    def requests(self):

        yield _ModelGroupPredict(
            model_id=self.model_id, data=self.data, predict_options=self.predict_options
        )


class UpdateModelGroupSettings(GraphQLRequest):
    """
    Updates an existing model group component in the platform.

    Args:
        model_group_id (int): the id of the model group to update settings
        model_training_options (dict, optional): model training options to use when training model. Defaults to None
            Valid options are based on model type:
            - Text Extraction: 'max_empty_chunk_ratio', 'auto_negative_scaling', 'optimize_for', 'subtoken_prediction', 'base_model', 'class_weight'
            - Text Classification: 'model_type'
            - Object Detection / Image Classification: 'filter_empty', 'n_epochs', 'use_small_model'
        predict_options (dict, optional): predict options to use on model. Defaults to None
            Valid options are based on model type:
            - Object Detection: 'threshold', 'predict_batch_size'
            - Finetune: 'negative_confidence'
            - Document: 'negative_confidence'
    """

    query = """
        mutation updateModelGroup(
            $modelGroupId: Int!,
            $modelTrainingOptions: JSONString,
            $predictOptions: JSONString,
        ) {
            updateModelGroupSettings(
                modelGroupId: $modelGroupId,
                modelTrainingOptions: $modelTrainingOptions,
                predictOptions: $predictOptions,
            ) {
                modelOptions {
                    id
                    domain
                    highQuality
                    interlabelerResolution
                    samplingStrategy
                    seed
                    testSplit
                    weightByClassFrequency
                    wordPredictorStrength
                    modelTrainingOptions
                    predictOptions
                }
            }
        }
    """

    def __init__(
        self,
        model_group_id: int,
        model_training_options: Optional[Dict[str, Any]] = None,
        predict_options: Optional[Dict[str, Any]] = None,
    ):
        if model_training_options:
            model_training_options = json.dumps(model_training_options)

        if predict_options:
            predict_options = json.dumps(predict_options)

        super().__init__(
            self.query,
            variables={
                "modelGroupId": model_group_id,
                "modelTrainingOptions": model_training_options,
                "predictOptions": predict_options,
            },
        )

    def process_response(self, response):
        return ModelOptions(
            **super().process_response(response)["updateModelGroupSettings"][
                "modelOptions"
            ]
        )



import io
import warnings
from typing import List, Union

from indico.client import Delay, GraphQLRequest, RequestChain
from indico.errors import IndicoRequestError
from indico.queries.storage import RetrieveStorageObject
from indico.types.export import Export, LabelResolutionStrategy


class _CreateExport(GraphQLRequest):
    query = """
        mutation CreateExport(
            $datasetId: Int!,
            $labelsetId: Int!,
            $columnIds: [Int],
            $modelIds: [Int],
            $frozenLabelsetIds: [Int],
            $combineLabels: LabelResolutionStrategy,
            $fileInfo: Boolean,
            $anonymous: Boolean
        ) {
            createExport(
                datasetId: $datasetId,
                labelsetId: $labelsetId,
                columnIds: $columnIds,
                modelIds: $modelIds,
                frozenLabelsetIds: $frozenLabelsetIds,
                combineLabels: $combineLabels,
                fileInfo: $fileInfo,
                anonymous: $anonymous
            ) {
                id,
                datasetId,
                labelsetId,
                name,
                status,
                columnIds,
                modelIds,
                frozenLabelsetIds,
                anonymous
                downloadUrl,
            }
        }

    """

    def __init__(
        self,
        dataset_id: int,
        labelset_id: int,
        column_ids: List[int] = None,
        model_ids: List[int] = None,
        frozen_labelset_ids: List[int] = None,
        combine_labels: LabelResolutionStrategy = LabelResolutionStrategy.ALL.name,
        file_info: bool = None,
        anonymoous: bool = None,
        anonymous: bool = None,
    ):
        if anonymoous:
            warnings.warn(
                "Argument anonymoous is deprecated and will be removed in future versions. Use argument anonymous instead."
            )
            if anonymous:
                raise IndicoRequestError("Cannot use both anonymoous and anonymous.")
            else:
                anonymous = anonymoous
        super().__init__(
            self.query,
            variables={
                "datasetId": dataset_id,
                "labelsetId": labelset_id,
                "columnIds": column_ids,
                "modelIds": model_ids,
                "frozenLabelsetIds": frozen_labelset_ids,
                "combineLabels": combine_labels,
                "fileInfo": file_info,
                "anonymous": anonymous,
            },
        )

    def process_response(self, response):
        response = super().process_response(response)
        return Export(**response["createExport"])


class GetExport(GraphQLRequest):
    """
    Get information on an Export job

    Args:
        export_id (int): Id of an Export

    Returns:
        Export object

    """

    query = """
        query GetExportStatus($exportIds: [Int]) {
            exports (exportIds: $exportIds) {
                exports {
                id
                datasetId
                name
                status
                columnIds
                labelsetId
                modelIds
                frozenLabelsetIds
                numLabels
                anonymous
                downloadUrl
                createdAt
                }
            }
        }
    """

    def __init__(self, export_id: int):
        super().__init__(self.query, variables={"exportIds": [export_id]})

    def process_response(self, response):
        response = super().process_response(response)
        return Export(**response["exports"]["exports"][0])


class _RetrieveExport(RetrieveStorageObject):
    def process_response(self, response):
        try:
            import pandas as pd
        except ImportError as error:
            raise RuntimeError(
                "downloading exports requires additional dependencies: "
                "`pip install indico-client[exports]`"
            ) from error

        response = super().process_response(response)
        return pd.read_csv(io.StringIO(response))


class DownloadExport(RequestChain):
    """
    Download an export from an Indico storage url

    Args:
        export (Export): Export object
        export_id (int): Export id

    Returns:
        Pandas csv of the export

    Raises:
        IndicoRequestError if the Export job is not complete or failed
    """

    def __init__(self, export_id: int = None, export: Export = None):
        if not export_id and not export:
            raise IndicoRequestError(
                code="FAILURE",
                error="Please provide at least one of <export_id> or <export>",
            )
        self.export_id = export_id
        self.export = export

    def requests(self):
        if self.export_id:
            yield GetExport(self.export_id)
        export = self.export or self.previous

        if export.status != "COMPLETE":
            raise IndicoRequestError(
                code="400",
                error="The export must be complete and not failed before it can be downloaded",
            )

        yield _RetrieveExport(export.download_url)


class CreateExport(RequestChain):
    """
    Create an export job for a dataset.

    Args:
        dataset_id (int): Dataset to create the export for.
        labelset_id (int): Labelset column id to export.
        column_ids (List(int), optional): Data column ids to export. Defaults to None.
        model_ids (List(int), optional): Model ids to include predictions from. Defaults to None.
        frozen_labelset_ids: (List(int), optional): frozen labelset ids to limit examples by. Defaults to None.
        combine_labels (LabelResolutionStrategy, optional): One row per example, combine labels from multiple labels into a single row. Defaults to 'all'.
        file_info (bool, optional): Include datafile information. Defaults to False.
        anonymous (bool, optional): Anonymize user information. Defaults to False.
        wait (bool, optional): Wait for the export to complete. Defaults to True.
        request_interval (int or float, optional): The maximum time in between retry calls when waiting. Defaults to 5 seconds.

    Returns:
        Export object

    """

    previous = None

    def __init__(
        self,
        dataset_id: int,
        labelset_id: int,
        column_ids: List[int] = None,
        model_ids: List[int] = None,
        frozen_labelset_ids: List[int] = None,
        combine_labels: LabelResolutionStrategy = LabelResolutionStrategy.ALL.name,
        file_info: bool = False,
        anonymous: bool = False,
        wait: bool = True,
        request_interval: Union[int, float] = 5,
    ):
        self.dataset_id = dataset_id
        self.labelset_id = labelset_id
        self.column_ids = column_ids
        self.model_ids = model_ids
        self.frozen_labelset_ids = frozen_labelset_ids
        self.combine_labels = combine_labels
        self.file_info = file_info
        self.anonymous = anonymous
        self.wait = wait
        self.request_interval = request_interval
        super().__init__()

    def requests(self):
        yield _CreateExport(
            dataset_id=self.dataset_id,
            labelset_id=self.labelset_id,
            column_ids=self.column_ids,
            model_ids=self.model_ids,
            frozen_labelset_ids=self.frozen_labelset_ids,
            combine_labels=self.combine_labels,
            file_info=self.file_info,
            anonymous=self.anonymous,
        )
        if self.wait is True:
            while self.previous.status not in ["COMPLETE", "FAILED"]:
                yield GetExport(self.previous.id)
                yield Delay(seconds=self.request_interval)

        yield GetExport(self.previous.id)



from typing import List

from indico.types import BaseType
from indico.types import SubmissionFile


class DocumentReport(BaseType):
    """
    A report about a submission

    """
    dataset_id: int
    workflow_id: int
    submission_id: int
    status: str
    created_at: str
    created_by: str
    updated_at: str
    updated_by: str
    completed_at: str
    errors: str
    retrieved: bool
    input_files: List[SubmissionFile]
    files_deleted: bool




import os
import time
from datetime import datetime
from pathlib import Path

import pytest

from indico.client import IndicoClient
from indico.errors import IndicoError, IndicoInputError, IndicoTimeoutError
from indico.filters import SubmissionFilter, SubmissionReviewFilter
from indico.queries import (
    AddDataToWorkflow,
    AddFiles,
    CreateDataset,
    DeleteWorkflow,
    GetSubmission,
    JobStatus,
    ListSubmissions,
    ListWorkflows,
    RetrieveStorageObject,
    SubmissionResult,
    SubmitReview,
    UpdateSubmission,
    UpdateWorkflowSettings,
    WaitForSubmissions,
    WorkflowSubmission,
    WorkflowSubmissionDetailed,
)
from indico.queries.questionnaire import GetQuestionnaire
from indico.queries.workflow import GetWorkflow
from indico.queries.workflow_components import _AddWorkflowComponent
from indico.types import ModelGroup, ModelTaskType, NewLabelsetArguments
from indico.types.submission import Submission

from ..data.datasets import *  # noqa
from ..data.datasets import PUBLIC_URL


def test_list_workflows(indico, airlines_dataset, airlines_model_group: ModelGroup):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    assert len(wfs) > 0


def test_list_workflows_audit_info(
    indico, airlines_dataset, airlines_model_group: ModelGroup
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    assert wfs[0].created_at
    assert isinstance(wfs[0].created_at, datetime)
    assert wfs[0].created_by
    assert isinstance(wfs[0].created_by, int)


@pytest.mark.parametrize(
    "_input",
    [
        {"urls": [PUBLIC_URL + "mock.pdf"]},
        {"files": [str(Path(__file__).parents[1]) + "/data/mock.pdf"]},
    ],
)
def test_workflow_submission(
    indico, airlines_dataset, airlines_model_group: ModelGroup, _input
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submission_ids = client.call(WorkflowSubmission(workflow_id=wf.id, **_input))
    submission_id = submission_ids[0]
    assert submission_id is not None

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "FAILED"))

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "INVALID_STATUS"))

    result_url = client.call(
        SubmissionResult(submission_id, "COMPLETE", wait=True, timeout=120)
    )
    result = client.call(RetrieveStorageObject(result_url.result))
    assert isinstance(result, dict)
    assert result["submission_id"] == submission_id
    assert result["file_version"] == 1
    sub = client.call(GetSubmission(submission_id))
    assert isinstance(sub, Submission)
    assert sub.retrieved is False
    assert sub.files_deleted is False
    client.call(UpdateSubmission(submission_id, retrieved=True))
    sub = client.call(GetSubmission(submission_id))
    assert sub.retrieved is True


def test_workflow_submission_with_streams(
    indico, airlines_dataset, airlines_model_group: ModelGroup
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    path = Path(str(Path(__file__).parents[1]) + "/data/mock.pdf")
    fd = open(path.absolute(), "rb")
    files = {"mock.pdf": fd}
    submission_ids = client.call(WorkflowSubmission(workflow_id=wf.id, streams=files))
    submission_id = submission_ids[0]
    assert submission_id is not None

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "FAILED"))

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "INVALID_STATUS"))

    result_url = client.call(
        SubmissionResult(submission_id, "COMPLETE", wait=True, timeout=120)
    )
    result = client.call(RetrieveStorageObject(result_url.result))
    assert isinstance(result, dict)
    assert result["submission_id"] == submission_id
    assert result["file_version"] == 1
    client.call(UpdateSubmission(submission_id, retrieved=True))
    sub = client.call(GetSubmission(submission_id))
    assert isinstance(sub, Submission)
    assert sub.retrieved is True


def test_workflow_submission_with_text(
    indico, airlines_dataset, airlines_model_group: ModelGroup
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submission_ids = client.call(
        WorkflowSubmission(workflow_id=wf.id, text="hello this is a test")
    )
    submission_id = submission_ids[0]
    assert submission_id is not None

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "FAILED"))

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "INVALID_STATUS"))

    result_url = client.call(
        SubmissionResult(submission_id, "COMPLETE", wait=True, timeout=120)
    )
    result = client.call(RetrieveStorageObject(result_url.result))
    assert isinstance(result, dict)
    assert result["submission_id"] == submission_id
    assert result["file_version"] == 1
    client.call(UpdateSubmission(submission_id, retrieved=True))
    sub = client.call(GetSubmission(submission_id))
    assert isinstance(sub, Submission)
    assert sub.retrieved is True


@pytest.mark.parametrize(
    "_input",
    [
        {"urls": [PUBLIC_URL + "mock.pdf"] * 3},
        {"files": [str(Path(__file__).parents[1]) + "/data/mock.pdf"] * 3},
    ],
)
def test_workflow_submission_versioned(
    indico, airlines_dataset, airlines_model_group: ModelGroup, _input
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submission_ids = client.call(
        WorkflowSubmission(workflow_id=wf.id, result_version="LATEST", **_input)
    )

    assert len(submission_ids) == len(next(iter(_input.values())))
    submission_id = submission_ids[0]
    assert submission_id is not None

    submissions = client.call(WaitForSubmissions(submission_id, timeout=120))
    result = client.call(RetrieveStorageObject(submissions[0].result_file))

    assert isinstance(result, dict)
    assert result["file_version"] == 3
    assert len(result["submission_results"]) == 1
    assert (
        os.path.basename(result["submission_results"][0]["input_filename"])
        == "mock.pdf"
    )


@pytest.mark.parametrize(
    "_input",
    [
        {"urls": [PUBLIC_URL + "mock.pdf"] * 3},
        {"files": [str(Path(__file__).parents[1]) + "/data/mock.pdf"] * 3},
    ],
)
def test_workflow_submission_bundled(
    indico, airlines_dataset, airlines_model_group: ModelGroup, _input
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submission_ids = client.call(
        WorkflowSubmission(
            workflow_id=wf.id, bundle=True, result_version="LATEST", **_input
        )
    )

    assert len(submission_ids) == 1
    submission_id = submission_ids[0]
    assert submission_id

    submissions = client.call(WaitForSubmissions(submission_id, timeout=120))
    result = client.call(RetrieveStorageObject(submissions[0].result_file))

    assert isinstance(result, dict)
    assert result["file_version"] == 3
    assert len(result["submission_results"]) == len(next(iter(_input.values())))
    assert (
        os.path.basename(result["submission_results"][0]["input_filename"])
        == "mock.pdf"
    )


@pytest.mark.parametrize(
    "_input,_output",
    [
        ({"urls": [PUBLIC_URL + "mock.pdf"]}, PUBLIC_URL + "mock.pdf"),
        ({"files": [str(Path(__file__).parents[1]) + "/data/mock.pdf"]}, "mock.pdf"),
    ],
)
def test_workflow_submission_detailed(
    indico, airlines_dataset, airlines_model_group: ModelGroup, _input, _output
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submissions = client.call(WorkflowSubmissionDetailed(workflow_id=wf.id, **_input))
    assert isinstance(submissions[0], Submission)
    assert submissions[0].input_filename == _output


@pytest.mark.parametrize(
    "_input",
    [
        {"urls": [PUBLIC_URL + "mock.pdf"] * 3},
        {"files": [str(Path(__file__).parents[1]) + "/data/mock.pdf"] * 3},
    ],
)
def test_workflow_submission_timeout(
    indico, airlines_dataset, airlines_model_group: ModelGroup, _input
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submission_ids = client.call(
        WorkflowSubmission(
            workflow_id=wf.id, bundle=True, result_version="LATEST", **_input
        )
    )

    assert len(submission_ids) == 1
    submission_id = submission_ids[0]
    assert submission_id

    with pytest.raises(IndicoTimeoutError):
        start = time.time()
        client.call(WaitForSubmissions(submission_id, timeout=1))
        end = time.time()
        assert end - start >= 1 and end - start < 2

    with pytest.raises(IndicoTimeoutError):
        start = time.time()
        client.call(SubmissionResult(submission_id, "COMPLETE", wait=True, timeout=1))
        end = time.time()
        assert end - start >= 1 and end - start < 2


def test_list_workflow_submission_retrieved(
    indico, airlines_dataset, airlines_workflow, airlines_model_group: ModelGroup
):
    client = IndicoClient()

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    submission_ids = client.call(
        WorkflowSubmission(workflow_id=airlines_workflow.id, files=[dataset_filepath])
    )
    submission_id = submission_ids[0]
    assert submission_id is not None
    client.call(SubmissionResult(submission_id, "COMPLETE", wait=True))
    client.call(UpdateSubmission(submission_id, retrieved=True))

    submissions = client.call(
        ListSubmissions(filters=SubmissionFilter(retrieved=True, status="COMPLETE"))
    )
    assert all([s.retrieved for s in submissions])
    assert submission_id in [s.id for s in submissions]

    submissions = client.call(
        ListSubmissions(filters=SubmissionFilter(retrieved=False))
    )
    assert all([not s.retrieved for s in submissions])
    assert submission_id not in [s.id for s in submissions]


def test_list_workflow_submission_paginate(
    indico, airlines_dataset, airlines_model_group: ModelGroup
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    submission_ids = client.call(
        WorkflowSubmission(workflow_id=wf.id, files=[dataset_filepath] * 5)
    )
    subs = []

    for sub in client.paginate(ListSubmissions(workflow_ids=[wf.id], limit=3)):
        subs.extend(sub)
    for sub in subs:
        if not submission_ids:
            break
        assert sub.id == submission_ids.pop()  # list is desc by default


def test_workflow_submission_missing_workflow(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    with pytest.raises(IndicoError):
        client.call(WorkflowSubmission(workflow_id=0, files=[dataset_filepath]))


def test_workflow_submission_mixed_args(indico, airlines_dataset):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    _file = str(Path(__file__).parents[1]) + "/data/mock.pdf"
    url = PUBLIC_URL + "mock.pdf"

    with pytest.raises(IndicoInputError):
        client.call(WorkflowSubmission(workflow_id=wf.id, files=[_file], urls=[url]))


def test_auto_review_enabled_in_get_workflow_response(
    indico, org_annotate_dataset, org_annotate_model_group
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[org_annotate_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    assert not (wf.auto_review_enabled or wf.review_enabled)

    wf = client.call(
        UpdateWorkflowSettings(wf, enable_review=True, enable_auto_review=True)
    )
    assert wf.review_enabled and wf.auto_review_enabled

    wf = client.call(GetWorkflow(workflow_id=wf.id))
    assert wf.review_enabled and wf.auto_review_enabled


@pytest.mark.parametrize("force_complete", [None, True])
def test_workflow_submission_auto_review_v1(
    indico,
    force_complete,
    org_annotate_dataset,
    org_annotate_workflow,
    org_annotate_model_group,
):

    client = IndicoClient()

    wf = client.call(
        UpdateWorkflowSettings(
            org_annotate_workflow.id, enable_review=True, enable_auto_review=True
        )
    )
    assert wf.review_enabled and wf.auto_review_enabled

    _file = str(Path(__file__).parents[1]) + "/data/org-sample.pdf"

    sub_ids = client.call(
        WorkflowSubmission(workflow_id=wf.id, files=[_file], result_version="ONE")
    )
    subs = client.call(WaitForSubmissions(sub_ids, timeout=120))
    sub = subs[0]
    assert sub.status == "PENDING_AUTO_REVIEW"
    raw_result = client.call(RetrieveStorageObject(sub.result_file))
    changes = raw_result["results"]["document"]["results"]
    for model, preds in changes.items():
        if isinstance(preds, dict):
            preds["accepted"] = True
        elif isinstance(preds, list):
            for pred in preds:
                pred["accepted"] = True
    job = client.call(
        SubmitReview(sub.id, changes=changes, force_complete=force_complete)
    )
    job = client.call(JobStatus(job.id))
    submission = client.call(WaitForSubmissions([sub.id]))[0]
    assert (
        submission.status == "COMPLETE"
        if force_complete
        else submission.status == "PENDING_REVIEW"
    )


@pytest.mark.parametrize("force_complete", [None, True])
def test_workflow_submission_auto_review_v3_result(
    indico,
    force_complete,
    org_annotate_dataset,
    org_annotate_workflow,
    org_annotate_model_group,
):

    client = IndicoClient()

    wf = client.call(
        UpdateWorkflowSettings(
            org_annotate_workflow.id, enable_review=True, enable_auto_review=True
        )
    )
    assert wf.review_enabled and wf.auto_review_enabled

    _file = str(Path(__file__).parents[1]) + "/data/org-sample.pdf"

    sub_ids = client.call(
        WorkflowSubmission(workflow_id=wf.id, files=[_file], result_version="THREE")
    )
    subs = client.call(WaitForSubmissions(sub_ids, timeout=120))
    sub = subs[0]
    assert (
        sub.status == "PENDING_AUTO_REVIEW" or sub.status == "COMPLETE"
    )  # sub status will be set to COMPLETE if v3 is not supported
    if sub.status == "PENDING_AUTO_REVIEW":
        raw_result = client.call(RetrieveStorageObject(sub.result_file))
        changes = raw_result["submission_results"]
        assert isinstance(changes, list)
        for change in changes:
            # use original values
            change["model_results"] = change["model_results"]["ORIGINAL"]
            change["component_results"] = change["component_results"]["ORIGINAL"]
            for model, preds in change["model_results"].items():
                if isinstance(preds, dict):
                    preds["accepted"] = True
                elif isinstance(preds, list):
                    for pred in preds:
                        pred["accepted"] = True
        job = client.call(
            SubmitReview(sub.id, changes=changes, force_complete=force_complete)
        )
        job = client.call(JobStatus(job.id))
        submission = client.call(WaitForSubmissions([sub.id]))[0]
        assert (
            submission.status == "COMPLETE"
            if force_complete
            else submission.status == "PENDING_REVIEW"
        )


def test_list_workflow_submission_rejected(org_annotate_dataset):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[org_annotate_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)
    wf = client.call(
        UpdateWorkflowSettings(wf, enable_review=True, enable_auto_review=True)
    )
    assert wf.review_enabled and wf.auto_review_enabled

    _file = str(Path(__file__).parents[1]) + "/data/org-sample.pdf"

    sub_ids = client.call(WorkflowSubmission(workflow_id=wf.id, files=[_file]))
    subs = client.call(WaitForSubmissions(sub_ids, timeout=120))
    sub = subs[0]
    assert sub.status == "PENDING_AUTO_REVIEW"
    job = client.call(SubmitReview(sub.id, rejected=True))
    job = client.call(JobStatus(job.id))
    submissions = client.call(
        ListSubmissions(
            filters=SubmissionFilter(reviews=SubmissionReviewFilter(rejected=True))
        )
    )
    assert sub_ids[0] in [s.id for s in submissions]


def _new_dataset_for_updating(client):
    # new dataset
    airline_csv = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"
    dataset = client.call(
        CreateDataset(
            name=f"AddDataToWorkflow-test-{int(time.time())}", files=[airline_csv]
        )
    )

    workflowreq = CreateWorkflow(
        dataset_id=dataset.id, name=f"Update-test-{int(time.time())}"
    )
    wf = client.call(workflowreq)

    after_component_id = wf.component_by_type("INPUT_OCR_EXTRACTION").id
    source_col_id = dataset.datacolumn_by_name("Text").id
    classifier_name = f"CreateDatasetTeach-test-{int(time.time())}"

    new_labelset_args = {
        "datacolumn_id": source_col_id,
        "name": classifier_name,
        "num_labelers_required": 1,
        "task_type": ModelTaskType.CLASSIFICATION,
        "target_names": ["A", "B", "C"],
    }

    questionnaire = client.call(
        AddModelGroupComponent(
            name=classifier_name,
            dataset_id=dataset.id,
            workflow_id=wf.id,
            new_labelset_args=NewLabelsetArguments(**new_labelset_args),
            source_column_id=source_col_id,
            after_component_id=after_component_id,
        )
    )

    client.call(
        _AddWorkflowComponent(
            after_component_id=wf.component_by_type("OUTPUT_JSON_FORMATTER").id,
            component='{"component_type":"default_output","config":{}}',
            workflow_id=wf.id,
            after_component_link=None,
        )
    )

    questionnaire = client.call(
        GetQuestionnaire(
            questionnaire.model_group_by_name(
                classifier_name
            ).model_group.questionnaire_id
        )
    )

    # add data to dataset and process
    dataset = client.call(
        AddFiles(dataset_id=dataset.id, files=[airline_csv], autoprocess=True)
    )
    assert dataset.status == "COMPLETE"

    return dataset, wf, questionnaire


def test_add_data_to_workflow_wait(indico):
    client = IndicoClient()
    _, wf, q1 = _new_dataset_for_updating(client)

    wf = client.call(AddDataToWorkflow(wf.id, wait=True))
    assert wf.status == "COMPLETE"

    q2 = client.call(GetQuestionnaire(q1.id))

    # the total num available should double, since we re-add the same data
    assert q1.num_total_examples * 2 == q2.num_total_examples


def test_add_data_to_workflow_nowait(indico):
    client = IndicoClient()
    _, wf, _ = _new_dataset_for_updating(client)

    wf = client.call(AddDataToWorkflow(wf.id))
    assert wf.status == "ADDING_DATA"


@pytest.fixture
def wf_to_delete(indico, airlines_dataset: Dataset) -> Workflow:
    client = IndicoClient()
    workflowreq = CreateWorkflow(
        dataset_id=airlines_dataset.id,
        name=f"AirlineComplaints-test-{int(time.time())}",
    )
    response = client.call(workflowreq)

    return response


def test_delete_workflow(indico, airlines_dataset: Dataset, wf_to_delete: Workflow):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    num_wfs = len(wfs)
    client.call(DeleteWorkflow(workflow_id=wf_to_delete.id))
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    assert len(wfs) == num_wfs - 1
    assert wf_to_delete.id not in {wf.id for wf in wfs}
